from time import clock_gettime, CLOCK_REALTIME


from stl.formula import *
from stl.semantics import robustness_semantics


#
# Falsification
#
# Runs the falsification loop.  It takes as input a model (a function that maps
# inputs (positions or signals) to output signals, a formula to falsify, an
# optimisation method to find the next input to produce, and a timeout to stop
# the algorithm if it does not find a counterexample.
#


# [look_ahead]: function that maps a formula to the length of the output signal
# needed to compute the robustness (or the boolean) semantics.  The parameter
# [t] is only relevant if the formula contains unbounded Until operators.
def look_ahead(phi, t):
    if (isinstance(phi, Equality) or
         isinstance(phi, GTInequality)):
        return 0
    if isinstance(phi, Negate):
        return look_ahead(phi.subformulae[0], t)
    if isinstance(phi, Conjunction):
        return max([0] + [look_ahead(_, t) for _ in phi.subformulae])
    if isinstance(phi, BoundedUntil):
        return (max(look_ahead(phi.subformulae[0], t),
                      look_ahead(phi.subformulae[1], t)) + phi.upper)
    if isinstance(phi, Until):
        return (max(look_ahead(phi.subformulae[0], t),
                      look_ahead(phi.subformulae[1], t)) + t)
    raise Exception("Unknown formula constructor")


# [struct] is a global structure that will be used to find more information
# about the result of the falsification algorithm.
struct = {}


# [resume_falsification]: resumes the falsification algorithm with additional
# time.  This function is also called when falsification starts.
def resume_falsification(model, phi, optimisation, t=1E-4,
                         timeout=60):
    """Resumes an ongoing falsification with additional time.
       - model: a dictionary containing:
         - the model's name,
         - the model's variables (continuous and discrete),
         - a function to call the model,
       - phi: the formula to falsify,
       - optimisation: the name of the optimisation method,
       - t: the length "attributed" to unbounded Until constructors when \
         generating signals,
       - timeout: the time limit before the algorithm stops"""
    global struct

    start = clock_gettime(CLOCK_REALTIME)
    current = clock_gettime(CLOCK_REALTIME)
    while current - start <= timeout:
        p = optimisation.generate(struct)
        s = model["run"](p, struct["time_sim"])
        r = robustness_semantics(s, 0, phi)
        struct["tried"].append((p, r))
        optimisation.update(p, r, struct)
        if r < 0:
            struct["time"] = current - start
            return (p, r)
        if r < struct["best"][1]:
            struct["best"] = (p, r)
        current = clock_gettime(CLOCK_REALTIME)
    struct["time"] += current - start
    return struct["best"]


# [falsification]: starts falsification.
def falsification(model, phi, optimisation, t=1E-4, timeout=60):
    """Performs falsification.
       - model: a dictionary containing:
         - the model's name,
         - the model's variables (continuous and discrete),
         - a function to call the model,
       - phi: the formula to falsify,
       - optimisation: the name of the optimisation method,
       - t: the length "attributed" to unbounded Until constructors when \
         generating signals,
       - timeout: the time limit before the algorithm stops"""
    global struct

    # [struct] is the global structure that will be used troughout the
    # falsification algorithm.  It contains the following fields:
    # - [mode]: the name of the variable that describes the mode of the system
    #   (or None if there is no mode).
    # - [number-modes]: the number of modes of the system (or None).
    # - [variables]: the set of continuous variable names in the system.
    # - [invariant]: the invariant (zone of "real" values for the continuous
    #   variables of the model) of the system.
    #   Here, [invariant] takes a variable name and returns an interval, which
    #   is a restriction of hybrid systems.
    # - [optimisation]: the Optimisation object to use.
    # - [time_sim]: the time for which the system should be simulated.
    # - [best]: the current best input, and the robustness of the corresponding
    #   output.
    # - [tried]: the list of all tried inputs, and the robustnesses of the
    #   corresponding output.
    # - [time]: the time spent by the algorithm.
    t_sim = look_ahead(phi, t)
    struct = {"mode"            : model["mode"],
              "number-modes"    : (model["number-modes"]
                                   if model["mode"] != None
                                   else 1),
              "variables"       : model["variables"],
              "invariant"       : model["invariant"],
              "optimisation"    : optimisation,
              "time_sim"        : t_sim,
              "best"            : (None, float("inf")),
              "tried"           : [],
              "for-optimisation": {},
              "time"            : 0}
    optimisation.initialise(struct)

    return resume_falsification(model, phi, optimisation, t=t, timeout=timeout)


# [test]: function that tests the falsification algorithm with different
# optimisation methods.
def _test():
    from models.navigation import model, formula
    from stl.timed_signal import Signal
    from falsification.optimisation import random, random_biased, nelder_mead_random_biased

    def test_invariant(b, v):
        if v in ["x", "y"]:
            return (0, 1)
        if v in ["vx", "vy"]:
            return (float("-inf"), float("inf"))

        raise Exception("Unknown variable: {}".format(v))

    def test_model_run(x, t):
        return Signal(model(x["b"],
                            x["x"], x["y"],
                            x["vx"], x["vy"],
                            t + 1E-2),
                      variables = ("b", "x", "y", "vx", "vy"),
                      lower = 0,
                      upper = t,
                      delta = 1E-2)

    test_model = {"run": test_model_run,
                  "mode": "b",
                  "number-modes": 16,
                  "variables": {"x", "y", "vx", "vy"},
                  "invariant": test_invariant}

    timeout = 60

    print("Using random method for {} seconds:".format(timeout))
    (s,r) = falsification(test_model, formula, random, timeout=timeout)
    print("input = {}".format(s))
    print("robustness = {}".format(r))
    print("tries = {}".format(len(struct["tried"])))
    print("tries with mode 12 = {}".
          format(len(list(filter(lambda x: x[0]["b"] == 12, struct["tried"])))),
          end="\n\n")

    lambda_ = 0.2
    print("Using random_biased (lambda={}) method for {} seconds:".
          format(lambda_, timeout))
    (s,r) = falsification(test_model, formula, random_biased(lambda_),
                          timeout=timeout)
    print("input = {}".format(s))
    print("robustness = {}".format(r))
    print("tries = {}".format(len(struct["tried"])))
    print("tries with mode 12 = {}".
          format(len(list(filter(lambda x: x[0]["b"] == 12, struct["tried"])))),
          end="\n\n")

    alpha   = 1
    beta    = 0.5
    gamma   = 1
    epsilon = 1E-8
    lambda_ = 0.2
    print("Using Nelder-Mead (alpha={}, beta={}, gamma={}, epsilon={}), ".
          format(alpha, beta, gamma, epsilon),
          end="")
    print("random_biased (lambda={}) method for {} seconds:".
          format(lambda_, timeout))
    (s,r) = falsification(test_model, formula,
                          nelder_mead_random_biased(alpha, beta, gamma, epsilon,
                                                    lambda_),
                          timeout=timeout)
    print("input = {}".format(s))
    print("robustness = {}".format(r))
    print("tries = {}".format(len(struct["tried"])))
    print("tries with mode 12 = {}".
          format(len(list(filter(lambda x: x[0]["b"] == 12, struct["tried"])))),
          end="\n\n")


if __name__ == "__main__":
    _test()
