from random import uniform

from utils import insert_sorted_list
from falsification.point import Point


#
# Nelder-Mead
# From "A simplex method for function minimization"
# The Nelder-Mead method is a minimisation algorithm that keeps a simplex as its
# base structure, and that is driven towards the minimum by simple geometric
# operations.
#


# The parameters of the Nelder-Mead algorithm:
# - [alpha]: the reflection factor; the bigger, the farther points are reflected
#   through a face; 0 < alpha
# - [beta]: the contraction factor; the bigger, the closer to the face points
#   are contracted
# - [gamma]: the expansion factor; the bigger, the farther from the face points
#    are expanded
_alpha = 1
_beta  = 0.5
_gamma = 1

# [nelder_mead_loop] performs one loop of the Nelder-Mead algorithm, as
# described in the original article.
# The point with maximum (worst) value is reflected through the complementary
# face:
# - if this yields the a smaller (better) value than all vertices, try to expand
#   the point farther, and use that as your new vertex if it's better than other
#   vertices,
# - if this yields an average value (better than at least one vertex), then use
#   that as your new vertex,
# - if it's worse than all values, try contracting the point closer to the face:
#   - if this is better than at least one vertex, keep this as your new vertex,
#   - otherwise, contract the whole simplex towards the best vertex.
def nelder_mead_loop(p, f):
    n = len(p) - 1
    (p, h) = (p[:-1], p[-1])
    zero = {}
    for i in p[0][0].get_coordinates():
        zero[i] = 0
    centroid = Point(zero, lambda x: True)
    for x in p:
        centroid = centroid + x[0]
    centroid = centroid / n
    refl = (1 + _alpha) * centroid - _alpha * h[0]
    yr = f(refl)
    if yr < p[0][1]: ## the new point is the new minimum, so try expanding
        exp = (1 + _gamma) * refl - _gamma * centroid
        ye = f(exp)
        if ye < p[0][1]: ## expansion works
            insert_sorted_list((exp, ye), p, key = lambda x: x[1])
            return p
    if yr <= p[-1][1]: ## the new point is average
                       ## (or expansion failed)
        insert_sorted_list((refl, yr), p, key = lambda x: x[1])
        return p
    ymin = yr
    if yr > h[1]: ## the point is bigger than the maximum
        ymin = h[1]
        h = (refl, yr)
    contr = _beta * h[0] + (1 - _beta) * centroid
    yc = f(contr)
    if yc > ymin:
        p.append(h)
        p = list(map(lambda x: (  (x[0] + p[0][0]) / 2,
                                f((x[0] + p[0][0]) / 2)),
                       p))
        p.sort(key = lambda x: x[1])
        return p
    insert_sorted_list((contr, yc), p, key = lambda x: x[1])
    return p


# [nelder-mead]: a simplified version of the Nelder-Mead algorithm that performs
# [n] loops.  In the actual algorithm (and in the implementation for
# falsification, in optimisation.py), the algorithm stops when some quantity
# reaches below a certain threshold (and, in falsification, and new simplex is
# built if there is time left and the formula has not been falsified yet).
def nelder_mead(p, f, n):
    """Implements the Nelder-Mead method.
       Input: a list [p] of n+1 points of an n-dimensional space
       Output: a point [x] whose value is close to a local minimum"""
    p = list(map(lambda x: (x, f(x)), p))
    p.sort(key = lambda x: x[1])
    for i in range(n):
        p = nelder_mead_loop(p, f)

    return p


# A simple convex function.
def convex(p):
    xs = p.get_coordinates()
    x = xs["x"]
    y = xs["y"]
    return (x - 1) ** 2 + (y + 1) ** 2


# Rosenbrock's "banana" function.  Finding the minimum algorithmically is
# non-trivial (indeed, our naive Nelder-Mead algorithm sometimes gets it
# wrong!).
def banana(p):
    xs = p.get_coordinates()
    x = xs["x"]
    y = xs["y"]
    return 100 * (y - x **2) ** 2 + (1 - x) ** 2


def _test():
    print("testing banana function (minimum 0 in (1,1))")

    print("starting points:")
    # The parasiting "lambda x: True" is here to express that all variables are
    # continuous.
    p1 = Point({"x" : uniform(-10, 10), "y" : uniform(-10, 10)}, lambda x: True)
    p2 = Point({"x" : uniform(-10, 10), "y" : uniform(-10, 10)}, lambda x: True)
    p3 = Point({"x" : uniform(-10, 10), "y" : uniform(-10, 10)}, lambda x: True)
    print(p1)
    print(p2)
    print(p3)

    ps = [p1, p2, p3]
    res = nelder_mead(ps, banana, 1000)

    print("result:")
    print("point: {}\nvalue: {}\n".format(res[0][0], res[0][1]))

    print("testing convex function (minimum 0 in (1,-1))")

    print("starting points:")
    p1 = Point({"x" : uniform(-10, 10), "y" : uniform(-10, 10)}, lambda x: True)
    p2 = Point({"x" : uniform(-10, 10), "y" : uniform(-10, 10)}, lambda x: True)
    p3 = Point({"x" : uniform(-10, 10), "y" : uniform(-10, 10)}, lambda x: True)
    print(p1)
    print(p2)
    print(p3)

    ps = [p1, p2, p3]
    res = nelder_mead(ps, convex, 1000)

    print("result:")
    print("point: {}\nvalue: {}\n".format(res[0][0], res[0][1]))

if __name__ == "__main__":
    _test()
