class Point(object):

    # A point is implemented as a dictionary d = {dim: (type, val), ...} where:
    # - dim is an identifier for the dimension,
    # - type is a boolean: True for "continuous", False for "discrete",
    # - val is a numerical value.
    # The discrete part is considered to be a mode: it cannot be computed upon,
    # and operations are defined if and only if the modes are equal between the
    # points.

    # Two possible ways to initialise:
    # - using a dictionary d = {dim: (type, val), ...},
    # - using a dictionary d = {dim: val} and a function dim -> type.
    def __init__(self, xs, *f):
        if not isinstance(xs, dict):
            raise TypeError("Point should be initiated with a dictionary.")

        self.xs = {}
        if len(f) == 0:
            for i in xs:
                if not (isinstance(xs[i], tuple) and len(xs[i]) == 2 and
                        isinstance(xs[i][0],bool)):
                    error = "Point should be initiated with a dictionary of pairs "
                    error += "(bool,number)."
                    raise TypeError(error)
                self.xs[i] = xs[i]
        elif len(f) == 1:
            f = f[0]
            for i in xs:
                self.xs[i] = (f(i), xs[i])

    def get_coordinates(self):
        res = {}
        for i in self.xs:
            res[i] = self.xs[i][1]
        return res

    def __getitem__(self, i):
        return self.xs[i]

    def __setitem__(self, i, x):
        self.xs[i] = x

    def opposite(self):
        res = {}
        for i in self.xs:
            if self.xs[i][0]:
                res[i] = (self.xs[i][0],-self.xs[i][1])
            else:
                res[i] = self.xs[i]
        return Point(res)

    def __add__(self, p):
        xs = p.xs
        res = {}
        if self.xs.keys() != xs.keys():
            error = "Cannot add points with different dimensions: "
            error += "{} and {}.".format(self.xs.keys(), xs.keys())
            raise ValueError(error)
        for i in self.xs:
            if self.xs[i][0] != xs[i][0]:
                error = "Cannot add points that have different types for the "
                error += "same dimension {}.".format(i)
                raise ValueError(error)
            if self.xs[i][0]:
                res[i] = (self.xs[i][0], self.xs[i][1] + xs[i][1])
            else:
                if self.xs[i][1] != xs[i][1]:
                    error = "Cannot add points that have different discrete "
                    error += "dimensions: {}.".format(i)
                    raise ValueError(error)
                res[i] = self.xs[i]
        return Point(res)

    def __sub__(self, p):
        return self.__add__(p.opposite())

    def __rmul__(self, x):
        res = {}
        for i in self.xs:
            if self.xs[i][0]:
                res[i] = (self.xs[i][0], x * self.xs[i][1])
            else:
                res[i] = self.xs[i]
        return Point(res)

    def __truediv__(self, x):
        res = {}
        for i in self.xs:
            if self.xs[i][0]:
                res[i] = (self.xs[i][0], self.xs[i][1] / x)
            else:
                res[i] = self.xs[i]
        return Point(res)

    def __repr__(self):
        res = {}
        for i in self.xs:
            res[i] = self.xs[i][1]
            # res[i] = self.xs[i] # if you want to see whether the variables are
                                  # discrete or continuous
        return res.__repr__()

if __name__ == "__main__":
    x = Point({"x": 0.7, "y": 3.3},  lambda x: True)
    y = Point({"x": 1.1, "y": -4.5}, lambda x: True)

    print("x   =", x)
    print("y   =", y)
    print("x+y =", x+y)
    print("2.x =", 2*x)

    xq = Point({"q": 5, "x": 0.7, "y": 3.3},  lambda x: x != "q")
    yq = Point({"q": 5, "x": 1.1, "y": -4.5}, lambda x: x != "q")

    print("xq    =", xq)
    print("yq    =", yq)
    print("xq+yq =", xq+yq)
    print("2.xq  =", 2*xq)
