from math import sin, cos


from stl.formula import *
from stl.timed_signal import Signal
from utils import threshold

#
# Aircraft
# From "Monte-Carlo Techniques for Falsification of Temporal Properties of
# Non-Linear Hybrid Systems".
# Models an aircraft.

# The formula to falsify: if the speed is between 250 and 260 between 1 and 1.5
# seconds, them it cannot stay between 230 and 240 between 3 and 4 seconds.
formula = Imply(BA(In("x1", 250, 260), 1, 1.5),
                BA(~ In("x1", 230, 240), 3, 4))


# parameters:
# - x1: aircraft speed
# - x2: flight path angle
# - x3: altitude
# - u = (u1, u2): signal, where
#   - u1: thrust
#   - u2: angle of attack
def aircraft(x1, x2, x3, u, delta_sim=1E-4):

    # constants necessary to run the model
    g, S, rho, m = 9.81, 158, 0.3804, 74E3
    minx1, maxx1, minx2, maxx2, minx3, maxx3 = 200, 260, -10, 10, 120, 150
    minu1, maxu1, minu2, maxu2 = 34386, 53973, 0, 16
    B0, B1, B2 = 0.07351, -1.5E-3, 6.1E-4
    C0, C1 = 0.1667, 0.109

    x1 = threshold(x1, minx1, maxx1)
    x2 = threshold(x2, minx2, maxx2)
    x3 = threshold(x3, minx3, maxx3)

    def gen(x1, x2, x3, u):
        t = u.lower
        yield (x1, x2, x3)
        while t <= u.upper:
            u1, u2 = u(t)
            u1 = threshold(u1, minu1, maxu1)
            u2 = threshold(u2, minu2, maxu2)

            factor = S * rho / 2 / m

            dx1 = -factor * x1 ** 2 * (B0 + B1 * u2 + B2 * u2 ** 2)
            dx1 += u1 / m - g * sin(x2)

            dx2 = factor * C0 * x1 - g * cos(x2) / x1
            dx2 += factor * C1 * x1 * u2

            dx3 = x1 * sin(x2)

            x1 += dx1 * delta_sim
            x2 += dx2 * delta_sim
            x3 += dx3 * delta_sim
            t += delta_sim
            yield (x1, x2, x3)
    ge = gen(x1, x2, x3, u)
    vs = [Variable(_) for _ in ['x1', 'x2', 'x3']]
    return Signal(ge, vs, u.lower, u.upper, delta_sim, description="Aircraft model applied to " + str(u))


if __name__ == '__main__':
    from random import uniform

    u = Signal(lambda t: (40000, 8), description="simple control")
    u.plot()
    o = aircraft(uniform(200, 260), uniform(-10, 10), uniform(120, 150), u)
    o.plot()

