from stl.formula import *
from math import sin, cos, pi


#
# Navigation
# From "Linear Hybrid System Falsification Through Descent".
# Defines a benchmark of a boat drifting in currents.  The waters are divided
# into sections where the currents go in different directions.
#

# [formula]: the formula to falsify.  It states that, if the boat starts in a
# certain zone (in green in the drawing, if lauched as the main module), then it
# cannot reach another zone (in red in the same drawing).
formula = Imply(## Initial position
                And(Eq(Var("b"), Const(12)),
                    In("x", 0.2, 0.8),
                    In("y", 0.2, 0.8),
                    In("vx", -0.4, 0.4),
                    In("vy", -0.4, 0.4)),
                ## Final position
                BA(Not(And(Eq(Var("b"), Const(3)),
                           Lt(Add(Mul(Add(Var("x"), Const(-0.5)),
                                      Add(Var("x"), Const(-0.5))),
                                  Mul(Add(Var("y"), Const(-0.5)),
                                      Add(Var("y"), Const(-0.5)))),
                              Const(0.16),
                              strict=False))),
                   0,
                   20))


# attributes of the navigation class:
# - [b]: mode of the hybrid system, each mode corresponds to one zone of the
#   currents, in which the water flows in the same direction
# - [x]: horizontal position within the mode (0 <= x <= 1)
# - [y]: vertical position within the mode (0 <= y <= 1)
# - [vx]: horizontal speed
# - [vx]: vertical speed
class Navigation:
    def __init__(self, C, delta_sim=1E-4):
        self.delta_sim = delta_sim
        self.b  = 0
        self.x  = 0
        self.y  = 0
        self.vx = 0
        self.vy = 0
        self.u  = []
        for i in range(len(C[0])):
            l = []
            for j in range(len(C)):
                l.append((sin(pi * C[i][j] / 4),
                          cos(pi * C[i][j] / 4)))
            self.u.append(l)

    def __call__(self, b, x, y, vx, vy, t):
        self.b  = b
        self.x  = x
        self.y  = y
        self.vx = vx
        self.vy = vy
        i = 0
        yield (self.b, self.x, self.y, self.vx, self.vy)
        while i * self.delta_sim <= t:
            dx = self.vx
            dy = self.vy
            dvx = (0.1 * (self.vy + self.u[self.b // len(self.u)][self.b % len(self.u)][1]) -
                   1.2 * (self.vx - self.u[self.b // len(self.u)][self.b % len(self.u)][0]))
            dvy = (0.1 * (self.vx + self.u[self.b // len(self.u)][self.b % len(self.u)][0]) -
                   1.2 * (self.vy - self.u[self.b // len(self.u)][self.b % len(self.u)][1]))
            self.x  += dx  * self.delta_sim
            self.y  += dy  * self.delta_sim
            self.vx += dvx * self.delta_sim
            self.vy += dvy * self.delta_sim
            while self.x < 0 and self.b % len(self.u[0]) != 0:
                self.x += 1
                self.b -= 1
            while self.x > 1 and (self.b + 1) % len(self.u[0]) != 0:
                self.x -= 1
                self.b += 1
            while self.y < 0 and self.b >= len(self.u[0]):
                self.y += 1
                self.b -= len(self.u[0])
            while self.y > 1 and self.b < (len(self.u) - 1) * len(self.u[0]):
                self.y -= 1
                self.b += len(self.u[0])
            yield (self.b, self.x, self.y, self.vx, self.vy)
            i += 1


C = [[2, 2, 1, 1],
     [1, 2, 3, 6],
     [3, 6, 5, 6],
     [4, 2, 3, 4]]


model = Navigation(C, 1E-2)


if __name__ == "__main__":
    from random import uniform
    from models.graphics import *

    win = GraphWin(width=400, height=400)

    for i in range(3):
        p1 = Point(100 * (i+1), 0)
        p2 = Point(100 * (i+1), 400)
        line = Line(p1, p2)
        line.draw(win)

    for j in range(3):
        p1 = Point(0, 100 * (j+1))
        p2 = Point(400, 100 * (j+1))
        line = Line(p1, p2)
        line.draw(win)

    p1 = Point(20, 20)
    p2 = Point(80, 80)
    r = Rectangle(p1, p2)
    r.setFill("green")
    r.draw(win)

    pt = Point(350, 350)
    c = Circle(pt, 40)
    c.setFill("red")
    c.draw(win)

    x = model(12, uniform(0.2, 0.8), uniform(0.2, 0.8), uniform(-0.4, 0.4),
              uniform(-0.4, 0.4), 20 + 1E-2)

    for p in x:
        pt = Point(      100 * (p[0]  % 4 + p[1]),
                   400 - 100 * (p[0] // 4 + p[2]))
        pt.setFill("red")
        pt.draw(win)

    x = model(12, uniform(0.2, 0.8), uniform(0.2, 0.8), uniform(-0.4, 0.4),
              uniform(-0.4, 0.4), 20 + 1E-2)

    for p in x:
        pt = Point(      100 * (p[0]  % 4 + p[1]),
                   400 - 100 * (p[0] // 4 + p[2]))
        pt.setFill("blue")
        pt.draw(win)

    x = model(12, uniform(0.2, 0.8), uniform(0.2, 0.8), uniform(-0.4, 0.4),
              uniform(-0.4, 0.4), 20 + 1E-2)

    for p in x:
        pt = Point(      100 * (p[0]  % 4 + p[1]),
                   400 - 100 * (p[0] // 4 + p[2]))
        pt.setFill("purple")
        pt.draw(win)
