from utils import threshold
from stl.timed_signal import Signal
from stl.formula import *


#
# Rail
#
# Implements the model of a simple robot on a rail.  This robot can accelerate
# on a rail, with an acceleration in [-10, 10].
#


# [formula]: the formula to falsify.  This is falsified if the robot both leaves
# its initial position between 0 and 5 seconds and comes back to it between 5
# and 10 seconds.
formula = Or(BA(Var("x") < Const(8), 0, 5), BA(Var("x") > Const(2), 5, 10))


def rail(u, delta_sim=1E-4):
    minu, maxu = -10, 10

    # [u] is a signal that represents the acceleration of the robot.
    def gen(u):
        t = u.lower
        x, vx = 0, 0
        yield (x, vx)
        while t <= u.upper:
            ut = threshold(u(t), minu, maxu)
            ax = 0.1 * ut
            x += vx * delta_sim
            vx += ax * delta_sim
            t += delta_sim
            yield (x, vx)
    g = gen(u)
    vs = [Variable(_) for _ in ['x', 'vx']]
    return Signal(g, vs, u.lower, u.upper, delta_sim, description="Simple robot model applied to " + str(u))


def test_plotting():
    u = Signal(lambda t: (10,), upper=10, description="first control")
    o = rail(u)
    u.plot()
    o.plot()

    def coffee(t):
        if t < 3:
            return 10
        else:
            return -10

    u3 = Signal(coffee, upper=10, delta=1E-3, description="second control")
    u3.plot()
    o3 = rail(u3)
    o3.plot()


if __name__ == '__main__':
    from stl.semantics import robustness_semantics

    u = Signal(lambda t: (10,), upper=10, description="first control")
    o = rail(u)
    u.plot()
    o.plot()

    x = Var('x')
    x_returns = BoundedAlways(x > Const(2), 5, 10)
    print(robustness_semantics(o, 0, x_returns))

    x_leaves = BoundedAlways(x < Const(8), 0, 5)
    print(robustness_semantics(o, 0, x_leaves))

    print(robustness_semantics(o, 0, Disjunction(x_leaves, x_returns)))
