from math import sin, cos

from utils import threshold
from stl.formula import *
from stl.timed_signal import Signal


#
# Robot
# From "Two-Layered Falsification of Hybrid Systems Guided by Monte Carlo Tree
# Search".
# The model of a free-flying robot.
#


# The formula to falsify.  Since the formula is a negation, falsifying it is the
# same as finding a witness the negated formula is true.  In this case, it
# exhibits a way the robot from the initial position to a desired zone.
formula = ~ (BE(And(In("x", 3.9, 4.1), In("y", 3.9, 4.1)), 0, 5))


def robot(u, delta_sim=1E-4):
    minu, maxu = -10, 10

    # signal u = (u1, u2, u3, u4) where ui are the thrusts of four different
    # motors
    def gen(u):
        t = u.lower
        x, y, phi, vx, vy, vphi = 0, 0, 0, 0, 0, 0
        yield (x, y, vx, vy)
        while t <= u.upper:
            u1, u2, u3, u4 = u(t)
            u1 = threshold(u1, minu, maxu)
            u2 = threshold(u2, minu, maxu)
            u3 = threshold(u3, minu, maxu)
            u4 = threshold(u4, minu, maxu)

            ax = 0.1 * (u1 + u3) * cos(phi) - 0.1 * (u2 + u4) * sin(phi)
            ay = 0.1 * (u1 + u3) * sin(phi) + 0.1 * (u2 + u4) * cos(phi)
            aphi = 5 * (u1 + u3) / 12 - 5 * (u2 + u4) / 12

            x    += vx   * delta_sim
            y    += vy   * delta_sim
            phi  += vphi * delta_sim
            vx   += ax   * delta_sim
            vy   += ay   * delta_sim
            vphi += aphi * delta_sim

            t += delta_sim
            yield (x, y, vx, vy)
    g = gen(u)
    vs = [Variable(_) for _ in ['x', 'y', 'vx', 'vy']]
    return Signal(g, vs, u.lower, u.upper, delta_sim, description="Robot model applied to " + str(u))


def _test():
    y = Var('y')
    u = Signal(lambda t: (0.1, 0.11, -0.1, 0.09),
               description="simple control signal")
    o = robot(u)
    u.plot()
    o.plot(['x', 'y', 'vx', 'vy'])
    # tests whether the y-coordinate is always >= 0
    y_pos = BoundedAlways(y >= Const(0), 0, 1)
    # tests whether the y-coordinate eventually exceeds 0.01
    y_big = BoundedEventually(y > Const(0.01), 0, 1)
    print("Is the y-coordinate always bigger than 0? {!s}".format(boolean_semantics(o, 0, y_pos)))
    print("...by how much? {!s}".format(robustness_semantics(o, 0, y_pos)))
    print("Is the y-coordinate eventually bigger than 0.01? {!s}".format(boolean_semantics(o, 0, y_big)))
    print("...by how much? {!s}".format(robustness_semantics(o, 0, y_big)))

    u2 = Signal(lambda t: (sin(t), cos(t), -cos(t), -sin(t)), description='silly control')
    u2.plot()
    o2 = robot(u2)
    o2.plot()


if __name__ == '__main__':
    from stl.semantics import boolean_semantics, robustness_semantics

    _test()
