import random
from rl.gadget import *
from rl.gpi import StateValuation


def _get_seed():
    return numpy.random.randint(2 << 13)


class Agent:
    STATES = 'states'

    def __init__(self, seed=None, name="Bond"):
        self.seed = seed if seed else _get_seed()
        self.name = name
        self.prev_state, self.prev_action = None, None
        self.gadgets = []

    def start(self, state, action_set):
        if self.seed:
            random.seed(self.seed)
        self.prev_state, self.prev_action = None, None
        for gadget in self.gadgets:
            gadget.start()

    def policy(self, state, action_set):
        self.prev_state = state

    def step(self, past_reward, state, action_set):
        for gadget in self.gadgets:
            gadget.step(past_reward, self.prev_state, self.prev_action)

    def finish(self):
        for gadget in self.gadgets:
            gadget.finish()

    def __str__(self):
        return "Agent " + self.name


class RandomAgent(Agent):
    def __init__(self, seed=None, name="Ramon"):
        super(RandomAgent, self).__init__(seed, name)

    def start(self, state, action_set):
        super(RandomAgent, self).start(state, action_set)
        return self.policy(state, action_set)

    def policy(self, state, action_set):
        if not action_set:
            return None

        return random.choice(list(action_set))

    def step(self, past_reward, state, action_set):
        super(RandomAgent, self).step(past_reward, state, action_set)
        return self.policy(state, action_set)

    def finish(self):
        super(RandomAgent, self).finish()

    def __str__(self):
        return "Random" + str(super(RandomAgent, self))


class FairAgent(Agent):
    def __init__(self, seed=None, name="Fairy"):
        super(FairAgent, self).__init__(seed, name)
        self.gadgets = [CounterGadget()]
        self._counter_gadget = self.gadgets[0]

    def start(self, state, action_set):
        super(FairAgent, self).start(state, action_set)
        return self.policy(state, action_set)

    def policy(self, state, action_set):
        super(FairAgent, self).policy(state, action_set)
        least_actions = self._counter_gadget.least_common_actions(state, action_set)
        return random.choice(least_actions)

    def step(self, past_reward, state, action_set):
        super(FairAgent, self).step(past_reward, state, action_set)
        return self.policy(state, action_set)

    def finish(self):
        super(FairAgent, self).finish()

    def __str__(self):
        return "Fair" + str(super(FairAgent, self))


class GreedyAgent(Agent):
    def __init__(self, epsilon=0, min_sampling=1, seed=None, name="Greg"):
        super(GreedyAgent, self).__init__(seed, name)
        self.gadgets = [AverageRewardGadget()]
        self._avg_gadget = self.gadgets[0]
        self.epsilon = epsilon
        self.min_sampling = min_sampling

    def start(self, state, action_set):
        super(GreedyAgent, self).start(state, action_set)
        return self.policy(state, action_set)

    def policy(self, state, action_set):
        super(GreedyAgent, self).policy(state, action_set)
        rare_actions = self._avg_gadget.rare_actions(state, action_set, bound=self.min_sampling)
        if rare_actions:
            self.prev_action = random.choice(rare_actions)
        else:
            if random.random() < self.epsilon:
                self.prev_action = random.choice(action_set)
            else:
                self.prev_action = random.choice(self._avg_gadget.highest_average_reward(state, action_set))
        return self.prev_action

    def step(self, past_reward, state, action_set):
        super(GreedyAgent, self).step(past_reward, state, action_set)
        return self.policy(state, action_set)

    def finish(self):
        super(GreedyAgent, self).finish()

    def __str__(self):
        return "Greedy" + super(GreedyAgent, self).__str__()


class UpperConfidenceAgent(Agent):
    def __init__(self, c=1, min_sampling=1, seed=None, name="Amy"):
        super(UpperConfidenceAgent, self).__init__(seed, name)
        self.gadgets = [ConfidenceGadget(c)]
        self._conf_gadget = self.gadgets[0]
        self.min_sampling = min_sampling

    def start(self, state, action_set):
        super(UpperConfidenceAgent, self).start(state, action_set)
        return self.policy(state, action_set)

    def policy(self, state, action_set):
        super(UpperConfidenceAgent, self).policy(state, action_set)
        rare_actions = self._conf_gadget.rare_actions(state, action_set, bound=self.min_sampling)
        if rare_actions:
            self.prev_action = random.choice(rare_actions)
        else:
            self.prev_action = random.choice(self._conf_gadget.highest_confidence(state, action_set))
        return self.prev_action

    def step(self, past_reward, state, action_set):
        super(UpperConfidenceAgent, self).step(past_reward, state, action_set)
        return self.policy(state, action_set)

    def finish(self):
        super(UpperConfidenceAgent, self).finish()

    def __str__(self):
        return "Greedy" + super(UpperConfidenceAgent, self).__str__()


class RecencyAgent(Agent):
    def __init__(self, epsilon=0.05, min_sampling=1, alpha=0.5, seed=None, name="Reece"):
        super(RecencyAgent, self).__init__(seed, name)
        self.gadgets = [RecencyRewardGadget(alpha)]
        self._avg_gadget = self.gadgets[0]
        self.epsilon = epsilon
        self.min_sampling = min_sampling

    def start(self, state, action_set):
        super(RecencyAgent, self).start(state, action_set)
        return self.policy(state, action_set)

    def policy(self, state, action_set):
        super(RecencyAgent, self).policy(state, action_set)
        rare_actions = self._avg_gadget.rare_actions(state, action_set, bound=self.min_sampling)
        if rare_actions:
            self.prev_action = random.choice(rare_actions)
        else:
            if random.random() < self.epsilon:
                self.prev_action = random.choice(action_set)
            else:
                self.prev_action = random.choice(self._avg_gadget.highest_average_reward(state, action_set))
        return self.prev_action

    def step(self, past_reward, state, action_set):
        super(RecencyAgent, self).step(past_reward, state, action_set)
        return self.policy(state, action_set)

    def finish(self):
        super(RecencyAgent, self).finish()

    def __str__(self):
        return "Recency" + super(RecencyAgent, self).__str__()


class PolicyIterationAgent(Agent):
    def __init__(self, min_sampling=2, seed=None, name="Peter"):
        super(PolicyIterationAgent, self).__init__(seed, name)
        self.gadgets = [AverageRewardGadget(), MonteCarloMap()]
        self._valuation = StateValuation([])
        self._learned_policy = {}
        self.min_sampling = min_sampling

    def start(self, state, action_set):
        super(PolicyIterationAgent, self).start(state, action_set)
        return self.policy(state, action_set)

    def policy(self, state, action_set):
        super(PolicyIterationAgent, self).policy(state, action_set)
        rare_actions = self.gadgets[0].rare_actions(state, action_set, bound=self.min_sampling)
        if rare_actions:
            self.prev_action = rare_actions[0]
        else:
            self.prev_action = self._learned_policy[state]
        return self.prev_action

    def standard_policy(self):
        sa_dict = self.gadgets[0].seen_state_action()
        return {_: self.policy(_, sa_dict[_]) for _ in sa_dict}

    def step(self, past_reward, state, action_set):
        super(PolicyIterationAgent, self).step(past_reward, state, action_set)
        if not self._learned_policy:
            self._learned_policy = self.standard_policy()
        self._valuation.iterate(self.gadgets[0], self.gadgets[1], self._learned_policy, count=2)
        self._learned_policy = self._valuation.extract_policy(self.gadgets[0], self.gadgets[1])
        return self.policy(state, action_set)

    def finish(self):
        super(PolicyIterationAgent, self).finish()

    def __str__(self):
        return "PolicyIterator" + super(PolicyIterationAgent, self).__str__()


class StateAgnosticGreedyAgent(Agent):
    def __init__(self, epsilon=0, min_sampling=1, seed=None, name="Jennifer"):
        super(StateAgnosticGreedyAgent, self).__init__(seed, name)
        self.gadgets = [ActionRewardGadget()]
        self._avg_gadget = self.gadgets[0]
        self.epsilon = epsilon
        self.min_sampling = min_sampling

    def start(self, state, action_set):
        if self.seed:
            random.seed(self.seed)
        self.prev_state, self.prev_action = None, None
        return self.policy(state, action_set)

    def policy(self, state, action_set):
        super(StateAgnosticGreedyAgent, self).policy(state, action_set)
        rare_actions = self._avg_gadget.rare_actions(action_set, bound=self.min_sampling)
        if rare_actions:
            self.prev_action = random.choice(rare_actions)
        else:
            if random.random() < self.epsilon:
                self.prev_action = random.choice(action_set)
            else:
                self.prev_action = random.choice(self._avg_gadget.highest_average_reward(action_set))
        return self.prev_action

    def step(self, past_reward, state, action_set):
        super(StateAgnosticGreedyAgent, self).step(past_reward, state, action_set)
        return self.policy(state, action_set)

    def finish(self):
        super(StateAgnosticGreedyAgent, self).finish()

    def __str__(self):
        return "StateAgnosticGreedy" + super(StateAgnosticGreedyAgent, self).__str__()
