import numpy as np
import numbers
from copy import deepcopy

from stl.timed_signal import Signal
from stl.semantics import robustness_semantics
from rl.gadget import Stopwatch
from models.navigation import model as navmodel
from models.rail import rail


def _get_seed():
    return np.random.randint(2 << 13)


class Environment:
    TOTAL = 'total_reward'
    AVG = 'avg_reward'
    REWARD = 'reward'
    STATE = 'state'
    ACTION = 'action'
    ACTION_SET = 'actions'

    def __init__(self, delta=1, seed=None):
        self.history = None
        self.delta = delta
        self.seed = seed if seed else _get_seed()
        self.status = {}
        self._reset_rewards()
        self.vars = sorted(list(self.status.keys()))

    def _reset_rewards(self):
        self.status[Environment.TOTAL] = 0
        self.status[Environment.AVG] = 0
        self.status[Environment.REWARD] = 0
        self.status[Environment.ACTION] = None

    def restart(self):
        self._reset_rewards()
        self.vars = sorted(list(self.status.keys()))
        vals = tuple([self.status[_] for _ in self.vars])
        self.history = Signal([vals], lower=0, upper=0, variables=self.vars,
                              delta=self.delta, description=self.codename())

    def step(self, action):
        if isinstance(self.status[Environment.REWARD], numbers.Real):
            self.status[Environment.TOTAL] += (self.status[Environment.REWARD] *
                                               self.delta)
        self.status[Environment.ACTION] = action
        self.status[Environment.AVG] = (self.status[Environment.TOTAL] /
                                        (self.history.upper + self.delta))

        vars = sorted(list(self.status.keys()))
        if self.vars != vars:
            raise ValueError("Environment state type changing.")

        vals = tuple([self.status[_] for _ in vars])
        self.history >> vals

    def finish(self, plot=True, ret=False):
        cm = {Environment.REWARD: 'xkcd:lightblue'}
        if plot:
            self.history.plot(variables=[Environment.REWARD, Environment.AVG],
                              colormap=cm)
        if ret:
            return self.history

    def __str__(self):
        return 'Undescribed Environment ({!s})'.format(self.seed)

    def codename(self):
        return str(self)


class Bandit(Environment):
    ACTION_LETTERS = ['p', 'o', 'i', 'l', 'k', 'j', 'm', 'n', 'b', 'u', 'y',
                      't', 'h', 'g', 'f', 'v', 'c', 'x', 'z', 'd', 's', 'a',
                      'r', 'e', 'w', 'q']

    def __init__(self, arms=2, delta=1, seed=None):
        super(Bandit, self).__init__(delta=delta, seed=seed)

        self.arms = min(arms, 26)
        self.status[Environment.STATE] = '*'
        self.status[Environment.ACTION_SET] = Bandit.ACTION_LETTERS[:self.arms]
        np.random.seed(self.seed)
        self._means = {_: np.random.randint(5 * arms)
                       for _ in self.status[Environment.ACTION_SET]}

    def restart(self):
        super(Bandit, self).restart()
        return deepcopy(self.status)

    def step(self, action):
        if action in self.status[Environment.ACTION_SET]:
            self.status[Environment.REWARD] = round(self._means[action] +
                                                    np.random.standard_normal(),
                                                    2)
        else:
            self.status[Environment.REWARD] = None

        super(Bandit, self).step(action)
        return deepcopy(self.status)

    def finish(self, plot=True, ret=False):
        return super(Bandit, self).finish(plot, ret)

    def __str__(self):
        return ('David\'s Multi-armed bandit Environment (arms={!s})'.
                format(self.arms))

    def codename(self):
        return 'AHAB (a{!s})'.format(self.arms)


class SuttonBartoBandit(Environment):
    ACTION_LETTERS = ['p', 'o', 'i', 'l', 'k', 'j', 'm', 'n', 'b', 'u', 'y',
                      't', 'h', 'g', 'f', 'v', 'c', 'x', 'z', 'd', 's', 'a',
                      'r', 'e', 'w', 'q']

    def __init__(self, arms=2, delta=1, seed=None):
        super(SuttonBartoBandit, self).__init__(delta=delta, seed=seed)

        self.arms = min(arms, 26)
        self.status[Environment.STATE] = '*'
        self.status[Environment.ACTION_SET] = SuttonBartoBandit.ACTION_LETTERS[:self.arms]
        np.random.seed(self.seed)
        self._means = {_: np.random.standard_normal()
                       for _ in self.status[Environment.ACTION_SET]}

    def restart(self):
        super(SuttonBartoBandit, self).restart()
        return deepcopy(self.status)

    def step(self, action):
        if action in self.status[Environment.ACTION_SET]:
            self.status[Environment.REWARD] = round(self._means[action] +
                                                    np.random.standard_normal(),
                                                    2)
        else:
            self.status[Environment.REWARD] = None

        super(SuttonBartoBandit, self).step(action)
        return deepcopy(self.status)

    def finish(self, plot=True, ret=False):
        return super(SuttonBartoBandit, self).finish(plot, ret)

    def __str__(self):
        return ('Sutton-Barto Multi-armed bandit Environment (arms={!s})'.
                format(self.arms))

    def codename(self):
        return 'SBAHAB (a{!s})'.format(self.arms)


class MovingReward(Environment):
    ACTION_LETTERS = ['p', 'o', 'i', 'l', 'k', 'j', 'm', 'n', 'b', 'u', 'y',
                      't', 'h', 'g', 'f', 'v', 'c', 'x', 'z', 'd', 's', 'a',
                      'r', 'e', 'w', 'q']

    def __init__(self, arms=2, delta=1, seed=None):
        super(MovingReward, self).__init__(delta=delta, seed=seed)

        self.arms = min(arms, 26)
        self.status[Environment.STATE] = '*'
        self.status[Environment.ACTION_SET] = MovingReward.ACTION_LETTERS[:self.arms]
        self.stopwatch = Stopwatch()
        np.random.seed(self.seed)
        self._starts = {_: 10 * np.random.standard_normal()
                        for _ in self.status[Environment.ACTION_SET]}
        self._ends = {_: 10 * np.random.standard_normal()
                      for _ in self.status[Environment.ACTION_SET]}

    def restart(self):
        super(MovingReward, self).restart()
        self.stopwatch.start()
        return deepcopy(self.status)

    def step(self, action):
        self.stopwatch.step(0, 0, 0)
        if action in self.status[Environment.ACTION_SET]:
            start, end = self._starts[action], self._ends[action]
            base_reward = ((start - end) * 1.01 ** (-self.stopwatch.time()) +
                           end)
            self.status[Environment.REWARD] = round(base_reward +
                                                    np.random.standard_normal(),
                                                    2)
        else:
            self.status[Environment.REWARD] = None

        super(MovingReward, self).step(action)
        return deepcopy(self.status)

    def finish(self, plot=True, ret=False):
        self.stopwatch.finish()
        return super(MovingReward, self).finish(plot, ret)

    def __str__(self):
        return 'Moving Reward Environment (arms={!s})'.format(self.arms)

    def codename(self):
        return 'RVR (a{!s})'.format(self.arms)


class PlaygroundSlide(Environment):
    TOP, BOTTOM, LADDER = 'top', 'bottom', 'ladder'
    STAY, NEXT, PREV = 'stay', 'next', 'prev'

    def __init__(self, delta=1, seed=None):
        super(PlaygroundSlide, self).__init__(delta=delta, seed=seed)

        self.status[Environment.STATE] = PlaygroundSlide.TOP
        self.status[Environment.ACTION_SET] = [PlaygroundSlide.STAY,
                                               PlaygroundSlide.NEXT,
                                               PlaygroundSlide.PREV]

    def restart(self):
        super(PlaygroundSlide, self).restart()
        self.status[Environment.STATE] = PlaygroundSlide.TOP
        return deepcopy(self.status)

    def step(self, action):
        if action not in self.status[Environment.ACTION_SET]:
            return None

        state = self.status[Environment.STATE]
        if action == PlaygroundSlide.STAY:
            self.status[Environment.REWARD] = -1
        else:
            if state == PlaygroundSlide.TOP:
                if action == PlaygroundSlide.NEXT:
                    self.status[Environment.REWARD] = 100
                    self.status[Environment.STATE] = PlaygroundSlide.BOTTOM
                else:
                    self.status[Environment.REWARD] = -1
                    self.status[Environment.STATE] = PlaygroundSlide.LADDER
            elif state == PlaygroundSlide.BOTTOM:
                if action == PlaygroundSlide.NEXT:
                    self.status[Environment.REWARD] = -1
                    self.status[Environment.STATE] = PlaygroundSlide.LADDER
                else:
                    self.status[Environment.REWARD] = -50
                    self.status[Environment.STATE] = PlaygroundSlide.TOP
            elif state == PlaygroundSlide.LADDER:
                if action == PlaygroundSlide.NEXT:
                    self.status[Environment.REWARD] = -9
                    self.status[Environment.STATE] = PlaygroundSlide.TOP
                else:
                    self.status[Environment.REWARD] = -1
                    self.status[Environment.STATE] = PlaygroundSlide.BOTTOM

        super(PlaygroundSlide, self).step(action)
        return deepcopy(self.status)

    def finish(self, plot=True, ret=False):
        return super(PlaygroundSlide, self).finish(plot, ret)

    def __str__(self):
        return 'Playground Slide Environment'

    def codename(self):
        return 'SLD'


class SimpleRobotEnvironment(Environment):
    def __init__(self, formula):
        super(SimpleRobotEnvironment, self).__init__()
        self.formula = formula
        self.prev_robustness = None
        self._reset()

    def _set_valid_actions(self):
        actions = []
        state = self.status[Environment.STATE]
        for i in range(10):
            if state[i] < 10:
                actions.append(tuple(2 if i == j else 0 for j in range(10)))
            if state[i] > -10:
                actions.append(tuple(-2 if i == j else 0 for j in range(10)))
        self.status[Environment.ACTION_SET] = actions

    def _reset(self):
        self.status[Environment.STATE] = (0,) * 10
        self.prev_robustness = None
        self._set_valid_actions()

    def _robustness(self):
        u = Signal([0], upper=10, delta=1,
                   array_list=[list(self.status[Environment.STATE]) + [0]])
        x = rail(u, delta_sim=1E-1)
        return robustness_semantics(x, 0, self.formula)

    def _plot(self):
        u = Signal([0], upper=10, delta=1,
                   array_list=[list(self.status[Environment.STATE]) + [0]])
        x = rail(u, delta_sim=1E-1)
        x.plot()

    def restart(self):
        super(SimpleRobotEnvironment, self).restart()
        self._reset()
        return deepcopy(self.status)

    def step(self, action):
        if action not in self.status[Environment.ACTION_SET]:
            raise ValueError("Non-valid action given to Simple Robot.")
        if self.prev_robustness is None:
            self.prev_robustness = self._robustness()
            self.status[Environment.TOTAL] = - self.prev_robustness

        self.status[Environment.STATE] = tuple(np.array(self.status[Environment.STATE]) +
                                               np.array(action))
        self._set_valid_actions()

        rs = self._robustness()
        self.status[Environment.REWARD] = - rs + self.prev_robustness
        self.prev_robustness = rs

        super(SimpleRobotEnvironment, self).step(action)
        return deepcopy(self.status)

    def finish(self, plot=True, ret=False):
        return super(SimpleRobotEnvironment, self).finish(plot, ret)

    def __str__(self):
        return 'SimpleRobotEnvironment'

    def codename(self):
        return 'SRE'


class SimpleNavEnvironment(Environment):
    def __init__(self, formula):
        super(SimpleNavEnvironment, self).__init__()
        self.formula = formula
        self.prev_robustness = None
        self._reset()

    def _set_valid_actions(self):
        actions = []
        state = self.status[Environment.STATE]
        bounds = [[0.2, 0.8], [0.2, 0.8], [-0.4, 0.4], [-0.4, 0.4]]
        for i in range(4):
            if state[i] <= bounds[i][1] - 0.05:
                actions.append(tuple(0.05 if i == j else 0 for j in range(4)))
            if state[i] <= bounds[i][1] - 0.02:
                actions.append(tuple(0.02 if i == j else 0 for j in range(4)))
            if state[i] >= bounds[i][0] + 0.05:
                actions.append(tuple(-0.05 if i == j else 0 for j in range(4)))
            if state[i] >= bounds[i][0] + 0.02:
                actions.append(tuple(-0.02 if i == j else 0 for j in range(4)))
        self.status[Environment.ACTION_SET] = actions

    def _reset(self):
        self.status[Environment.STATE] = (0.5,0.5,0,0)
        self.prev_robustness = None
        self._set_valid_actions()

    def _robustness(self):
        u = Signal(navmodel(12, *(list(self.status[Environment.STATE]) + [20])),
              variables = ( "b", "x", "y", "vx", "vy" ),
              lower = 0,
              upper = 20,
              delta = 1E-2)
        return robustness_semantics(u, 0, self.formula)

    def restart(self):
        super(SimpleNavEnvironment, self).restart()
        self._reset()
        return deepcopy(self.status)

    def step(self, action):
        if action not in self.status[Environment.ACTION_SET]:
            raise ValueError("Non-valid action given to Simple Navigator.")
        if self.prev_robustness is None:
            self.prev_robustness = self._robustness()
            self.status[Environment.TOTAL] = - self.prev_robustness

        self.status[Environment.STATE] = tuple(np.array(self.status[Environment.STATE]) +
                                               np.array(action))
        self._set_valid_actions()

        rs = self._robustness()
        self.status[Environment.REWARD] = - rs + self.prev_robustness
        self.prev_robustness = rs

        super(SimpleNavEnvironment, self).step(action)
        return deepcopy(self.status)

    def finish(self, plot=True, ret=False):
        best_robustness = max(self.history.arrays[5][1:])  # very hacky, but projection isn't working.
        idx = self.history.arrays[5].index(best_robustness)
        state = self.history.arrays[4][idx]
        print("best robustness @ state: " + str(-best_robustness) + " @ " + str(state))

        return super(SimpleNavEnvironment, self).finish(plot, ret)

    def __str__(self):
        return 'SimpleNavEnvironment'

    def codename(self):
        return 'SNE'
