from collections import Counter
import numpy


class Gadget:
    def __init__(self):
        pass

    def start(self):
        pass

    def step(self, reward, state, action):
        pass

    def finish(self):
        pass

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "Gadget()"


class Stopwatch(Gadget):
    def __init__(self):
        super(Stopwatch, self).__init__()
        self._time = 0

    def start(self):
        super(Stopwatch, self).start()
        self._time = 0

    def step(self, reward, state, action):
        super(Stopwatch, self).step(reward, state, action)
        self._time += 1

    def finish(self):
        super(Stopwatch, self).finish()

    def time(self):
        return self._time

    def __repr__(self):
        return "Stopwatch()"


class CounterGadget(Stopwatch):
    def __init__(self):
        super(CounterGadget, self).__init__()
        self._state_visits = Counter()
        self._action_counters = {}

    def start(self):
        super(CounterGadget, self).start()
        self._state_visits = Counter()
        self._action_counters = {}

    def _check_new_state(self, state, action_set=None):
        if state not in self._action_counters:
            self._action_counters[state] = Counter()

        if action_set:
            action_counter = self._action_counters[state]
            for action in action_set:
                if action not in action_counter:
                    action_counter[action] = 0

    def step(self, reward, state, action):
        self._check_new_state(state)
        super(CounterGadget, self).step(reward, state, action)

        self._state_visits[state] += 1
        self._action_counters[state][action] += 1

    def finish(self):
        super(CounterGadget, self).finish()

    def seen_states(self):
        return list(self._state_visits.keys())

    def seen_state_action(self):
        return {_: list(self._action_counters[_]) for _ in self._action_counters.keys()}

    def state_count(self, state):
        self._check_new_state(state)
        return self._state_visits[state]

    def action_count(self, state, action):
        self._check_new_state(state)
        return self._action_counters[state][action]

    def least_common_actions(self, state, action_set):
        self._check_new_state(state, action_set)

        action_counter = self._action_counters[state]
        least_multiplicity = min(action_counter.values())
        return [_ for _ in action_counter.keys() if action_counter[_] == least_multiplicity]

    def rare_actions(self, state, action_set, bound=1):
        self._check_new_state(state, action_set)

        action_counter = self._action_counters[state]
        return [_ for _ in action_counter.keys() if action_counter[_] < bound]

    def confidence_bound(self, state, action):
        self._check_new_state(state, [action])

        if self.state_count(state) == 0:
            return 0  # what's the right value here?
        return numpy.sqrt(numpy.log(self.action_count(state, action)) / self.state_count(state))

    def __repr__(self):
        return "CounterGadget()"


class ActionRewardGadget(Gadget):
    def __init__(self, alpha=0.9):
        super(ActionRewardGadget, self).__init__()
        self.alpha = alpha
        self._average_rewards = {}
        self._action_counter = Counter()

    def start(self):
        super(ActionRewardGadget, self).start()
        self._average_rewards = {}
        self._action_counter = {}

    def _check_new_action(self, action_set=None):
        if action_set:
            for action in action_set:
                if action not in self._average_rewards:
                    self._average_rewards[action] = 0
                if action not in self._action_counter:
                    self._action_counter[action] = 0

    def step(self, reward, state, action):
        self._check_new_action([action])
        super(ActionRewardGadget, self).step(reward, state, action)

        self._action_counter[action] += 1
        reward_tracker = self._average_rewards
        reward_tracker[action] += self.alpha * (reward - reward_tracker[action])

    def finish(self):
        super(ActionRewardGadget, self).finish()

    def least_common_actions(self, action_set):
        self._check_new_action(action_set)

        action_counter = self._action_counter
        least_multiplicity = min(action_counter.values())
        return [_ for _ in action_counter.keys() if action_counter[_] == least_multiplicity]

    def rare_actions(self, action_set, bound=1):
        self._check_new_action(action_set)

        action_counter = self._action_counter
        return [_ for _ in action_set if action_counter[_] < bound]

    def average_reward(self, action):
        self._check_new_action([action])
        return self._average_rewards[action]

    def highest_average_reward(self, action_set):
        self._check_new_action(action_set)

        reward_tracker = {_: self._average_rewards[_] for _ in action_set}
        return [_ for _ in reward_tracker.keys() if reward_tracker[_] == max(reward_tracker.values())]

    def __repr__(self):
        return "ActionRewardGadget()"


class AverageRewardGadget(CounterGadget):
    def __init__(self):
        super(AverageRewardGadget, self).__init__()
        self._average_rewards = {}

    def start(self):
        super(AverageRewardGadget, self).start()
        self._average_rewards = {}

    def _check_new_state(self, state, action_set=None):
        super(AverageRewardGadget, self)._check_new_state(state, action_set)
        if state not in self._average_rewards:
            self._average_rewards[state] = {}

        if action_set:
            reward_tracker = self._average_rewards[state]
            for action in action_set:
                if action not in reward_tracker:
                    reward_tracker[action] = 0

    def step(self, reward, state, action):
        self._check_new_state(state, [action])
        super(AverageRewardGadget, self).step(reward, state, action)

        reward_tracker = self._average_rewards[state]
        reward_tracker[action] += (reward - reward_tracker[action]) / self.action_count(state, action)

    def finish(self):
        super(AverageRewardGadget, self).finish()

    def average_reward(self, state, action):
        self._check_new_state(state, [action])
        return self._average_rewards[state][action]

    def highest_average_reward(self, state, action_set):
        self._check_new_state(state, action_set)

        reward_tracker = self._average_rewards[state]
        return [_ for _ in reward_tracker.keys() if reward_tracker[_] == max(reward_tracker.values())]

    def __repr__(self):
        return "AverageRewardGadget()"


class ConfidenceGadget(AverageRewardGadget):
    def __init__(self, c=1):
        super(ConfidenceGadget, self).__init__()
        self._confidence_scores = {}
        self.c = c

    def start(self):
        super(ConfidenceGadget, self).start()
        self._confidence_scores = {}

    def _check_new_state(self, state, action_set=None):
        super(ConfidenceGadget, self)._check_new_state(state, action_set)
        if state not in self._confidence_scores:
            self._confidence_scores[state] = {}

        if action_set:
            confidence_tracker = self._confidence_scores[state]
            for action in action_set:
                if action not in confidence_tracker:
                    confidence_tracker[action] = 0

    def step(self, reward, state, action):
        self._check_new_state(state, [action])
        super(ConfidenceGadget, self).step(reward, state, action)

        confidence_tracker = self._confidence_scores[state]
        confidence_tracker[action] = self.average_reward(state, action) + self.c * self.confidence_bound(state, action)

    def finish(self):
        super(ConfidenceGadget, self).finish()

    def confidence(self, state, action):
        self._check_new_state(state, [action])
        return self._confidence_scores[state][action]

    def highest_confidence(self, state, action_set):
        self._check_new_state(state, action_set)

        confidence_tracker = self._confidence_scores[state]
        return [_ for _ in confidence_tracker.keys() if confidence_tracker[_] == max(confidence_tracker.values())]

    def __repr__(self):
        return "ConfidenceGadget()"


class RecencyRewardGadget(CounterGadget):
    def __init__(self, alpha=0.5):
        super(RecencyRewardGadget, self).__init__()
        self._recent_rewards = {}
        self.alpha = alpha

    def start(self):
        super(RecencyRewardGadget, self).start()
        self._recent_rewards = {}

    def _check_new_state(self, state, action_set=None):
        super(RecencyRewardGadget, self)._check_new_state(state, action_set)
        if state not in self._recent_rewards:
            self._recent_rewards[state] = {}

        if action_set:
            reward_tracker = self._recent_rewards[state]
            for action in action_set:
                if action not in reward_tracker:
                    reward_tracker[action] = 0

    def step(self, reward, state, action):
        self._check_new_state(state, [action])
        super(RecencyRewardGadget, self).step(reward, state, action)

        reward_tracker = self._recent_rewards[state]
        reward_tracker[action] += self.alpha * (reward - reward_tracker[action])

    def finish(self):
        super(RecencyRewardGadget, self).finish()

    def average_reward(self, state, action):
        self._check_new_state(state, [action])
        return self._recent_rewards[state][action]

    def highest_average_reward(self, state, action_set):
        self._check_new_state(state, action_set)

        reward_tracker = self._recent_rewards[state]
        return [_ for _ in reward_tracker.keys() if reward_tracker[_] == max(reward_tracker.values())]

    def __repr__(self):
        return "RecencyRewardGadget()"


class MonteCarloMap(CounterGadget):
    def __init__(self):
        super(MonteCarloMap, self).__init__()
        self._transition_map = {}  # tau: S --> A --> S' --> (R, N)
        self._prevprev_state, self._prevprev_action = None, None
        # counts number of times this transition has been taken and pr(transition to s' | s, a)

    def start(self):
        super(MonteCarloMap, self).start()
        self._transition_map = {}
        self._prevprev_state, self._prevprev_action = None, None

    def _check_new_state(self, state, action_set=None, pp_state=None, pp_action=None):
        super(MonteCarloMap, self)._check_new_state(state, action_set)
        if state not in self._transition_map:
            self._transition_map[state] = {}

        if action_set:
            transition_tracker = self._transition_map[state]
            for action in action_set:
                if action not in transition_tracker:
                    transition_tracker[action] = {}

        if pp_state and pp_action:
            transition_tracker = self._transition_map[pp_state][pp_action]
            if state not in transition_tracker:
                transition_tracker[state] = [0, 0]

    def step(self, reward, state, action):
        self._check_new_state(state, [action], pp_state=self._prevprev_state, pp_action=self._prevprev_action)
        super(MonteCarloMap, self).step(reward, state, action)

        if self._prevprev_action and self._prevprev_state:
            transition_tracker = self._transition_map[self._prevprev_state][self._prevprev_action]
            transition_tracker[state][1] += 1

            state_action_total = self.action_count(self._prevprev_state, self._prevprev_action)
            # This total might be too high if our super class counted the current action in this set too...
            if state == self._prevprev_state and action == self._prevprev_action:
                state_action_total -= 1
            # Now we can safely update the distribution for the pp_state and pp_action
            for next_state in transition_tracker.keys():
                transition_tracker[next_state][0] = transition_tracker[next_state][1] / state_action_total

        self._prevprev_state = state
        self._prevprev_action = action

    def finish(self):
        super(MonteCarloMap, self).finish()

    def transition_probability(self, state, action, next_state):
        self._check_new_state(state, [action])
        self._check_new_state(next_state, [], pp_state=state, pp_action=action)
        return self._transition_map[state][action][next_state][0]

    def transition_count(self, state, action, next_state):
        self._check_new_state(state, [action])
        self._check_new_state(next_state, [], pp_state=state, pp_action=action)
        return self._transition_map[state][action][next_state][1]

    def transition_distribution(self, state, action):
        self._check_new_state(state, [action])
        transition_tracker = self._transition_map[state][action]
        return {_: transition_tracker[_][0] for _ in transition_tracker.keys() if transition_tracker[_][1] > 0}

    def possible_next_states(self, state, action):
        self._check_new_state(state, [action])
        transition_tracker = self._transition_map[state][action]
        return [_ for _ in transition_tracker.keys() if transition_tracker[_][1] > 0]

    def most_likely_next_state(self, state, action):
        self._check_new_state(state, [action])
        transition_tracker = self._transition_map[state][action]
        highest_count = max([_[1] for _ in transition_tracker.values()])
        return [_ for _ in transition_tracker.keys() if transition_tracker[_][1] == highest_count]

    def __repr__(self):
        return "MonteCarloMap()"


if __name__ == '__main__':
    mcm = MonteCarloMap()
    mcm.start()
    print('loading two state cycle...')
    mcm.step(0, 'ls', 'la')
    mcm.step(0, 'ls', 'ra')
    mcm.step(0, 'rs', 'ra')
    mcm.step(0, 'rs', 'la')
    mcm.step(0, 'ls', 'ra')

    print("ls -la-> : " + str(mcm.transition_distribution('ls', 'la')))
    print("ls -ra-> : " + str(mcm.transition_distribution('ls', 'ra')))
    print("rs -la-> : " + str(mcm.transition_distribution('rs', 'la')))
    print("rs -ra-> : " + str(mcm.transition_distribution('rs', 'ra')))

    print('\nAdding unexpected ls -ra-> ls transition...')
    mcm.step(0, 'ls', 'ra')

    print("ls -la-> : " + str(mcm.transition_distribution('ls', 'la')))
    print("ls -ra-> : " + str(mcm.transition_distribution('ls', 'ra')))
    print("rs -la-> : " + str(mcm.transition_distribution('rs', 'la')))
    print("rs -ra-> : " + str(mcm.transition_distribution('rs', 'ra')))

    print('\nQuerying for non-existing transition ls -la-> q...')
    mcm.transition_probability('ls', 'la', 'q')
    print("ls -la-> : " + str(mcm.transition_distribution('ls', 'la')))



