from rl.gadget import AverageRewardGadget, MonteCarloMap


class StateValuation:
    def __init__(self, state_set, default_value=0):
        self._map = {state: default_value for state in state_set}
        self._default_value = default_value

    def add_states(self, state_set):
        for state in state_set:
            if state not in self._map.keys():
                self._map[state] = self._default_value

    def _est_action_val(self, ars, mcm, state, action, gamma=0.9):
        transitions = mcm.transition_distribution(state, action)
        next_reward = sum([transitions[_] * self._map[_] for _ in transitions.keys()])
        return ars.average_reward(state, action) + gamma * next_reward

    def iterate_once(self, ars, mcm, policy, gamma=0.9):
        if not isinstance(ars, AverageRewardGadget) or not isinstance(mcm, MonteCarloMap):
            raise ValueError("Value iteration requires average reward and transition mappings.")
        self.add_states(ars.seen_states())
        self.add_states(mcm.seen_states())

        new_map = {}
        for state in self._map.keys():
            try:
                new_map[state] = self._est_action_val(ars, mcm, state, policy[state], gamma=gamma)
            except KeyError:
                new_map[state] = self._default_value
        max_change = max([abs(self._map[state] - new_map[state]) for state in self._map.keys()])
        self._map = new_map
        return max_change

    def iterate_for_tolerance(self, ars, mcm, policy, tolerance=0.001, gamma=0.9):
        max_change = self.iterate_once(ars, mcm, policy, gamma=gamma)
        while max_change > tolerance:
            max_change = self.iterate_once(ars, mcm, policy, gamma=gamma)

    def iterate(self, ars, mcm, policy, count=1, gamma=0.9):
        for _ in range(count):
            self.iterate_once(ars, mcm, policy, gamma=gamma)

    def extract_policy(self, ars, mcm, gamma=0.9):
        if not isinstance(ars, AverageRewardGadget) or not isinstance(mcm, MonteCarloMap):
            raise ValueError("Value iteration requires average reward and transition mappings.")
        self.add_states(ars.seen_states())
        self.add_states(mcm.seen_states())

        sa_dict = ars.seen_state_action()
        policy_dict = {}
        for state in sa_dict.keys():
            max_reward, recommended_action = float('-inf'), None
            for action in sa_dict[state]:
                reward = self._est_action_val(ars, mcm, state, action, gamma=gamma)
                if reward > max_reward:
                    max_reward = reward
                    recommended_action = action

            policy_dict[state] = recommended_action

        return policy_dict
