import rl.environment as e
import rl.agent as agt
from rl import part3 as skd
from stl.timed_signal import average_signals
from stl.formula import *
from models.navigation import formula as nav_formula


class ShellWrapper:
    def __init__(self, env):
        self.env = env
        user = 're'
        while user == 're':
            user = self.runloop()
        return

    @staticmethod
    def _get_input_in(msg, options):
        candidate = input(msg)
        while candidate not in options:
            print("Invalid input, please try again.")
            candidate = input(msg)
        return candidate

    def runloop(self):
        status = self.env.restart()
        prompt = "At {!s}, can do {!s}.\nAction? ".format(status['state'], status['actions'])
        valid_inputs = status['actions'] + ['re', 'st']
        user = ShellWrapper._get_input_in(prompt, valid_inputs)
        while user not in ['re', 'st']:
            status = self.env.step(user)
            prompt = "At {!s}, can do {!s}.\nAction? ".format(status['state'], status['actions'])
            if 'reward' in status:
                print("Reward: {!s}.\n".format(status['reward']))
            user = ShellWrapper._get_input_in(prompt, valid_inputs)
        self.env.finish()
        return user


def auto_run(environment, agent, plot=True, ret=False, steps=1000):
    t = 0
    status = environment.restart()
    action = agent.start(status[e.Environment.STATE], status[e.Environment.ACTION_SET])
    while t < steps:
        status = environment.step(action)
        action = agent.step(status[e.Environment.REWARD], status[e.Environment.STATE], status[e.Environment.ACTION_SET])
        t += 1
    agt_summary = agent.finish()
    print(agent.name + " finished, statistics to follow:")
    env_summary = environment.finish(plot, ret)
    return env_summary, agt_summary


def trial(environment, agent, steps=1000, repeat=100):
    t = 0
    avgs = []
    while t < repeat:
        report = auto_run(environment, agent, plot=False, ret=True, steps=steps)[0]
        report = report.project(variables=(e.Environment.TOTAL,))
        avgs.append(report)
        t += 1
    avg = average_signals(*avgs, description="total rewards for {} in {}.".format(agent.name, environment.codename()))
    return avg


def heat(environment, agent_list, steps=1000):
    results = []
    for agent in agent_list:
        result = auto_run(environment, agent, plot=False, ret=True, steps=steps)[0]
        result = result.project(variables=(e.Environment.TOTAL,))
        result.variables = [Variable(agent.name)]
        results.append(result)
    return results


def competition(environment_factory, agent_list, steps=1000, repeat=100):
    t = 0
    table = []
    while t < repeat:
        env = environment_factory()
        results = heat(env, agent_list, steps=steps)
        table.append(results)
        t += 1

    ret = None
    for agent_idx, agent in enumerate(agent_list):
        agent_results = [_[agent_idx] for _ in table]
        agent_avg = average_signals(*agent_results)
        agent_avg.variables = [Variable(agent.name)]
        if ret is None:
            ret = agent_avg
        else:
            ret = ret.pair(agent_avg)
    return ret


if __name__ == "__main__":
    r = agt.RandomAgent()
    g = agt.GreedyAgent(name='Greg Ayn')
    g2 = agt.GreedyAgent(epsilon=0.1, name="Greg Gecko")
    a = agt.UpperConfidenceAgent()
    a2 = agt.UpperConfidenceAgent(c=2, name="Amy2")
    r2 = agt.RecencyAgent()
    r3 = agt.RecencyAgent(alpha=0.9, name="Reese from the Past")
    r4 = agt.RecencyAgent(alpha=0.2, name="Reese from the Future")
    p = agt.PolicyIterationAgent()
    j = agt.StateAgnosticGreedyAgent(epsilon=0.1, min_sampling=5)
    sokendai = skd.CustomAgent()
    assignment_solution = skd.AssignmentSolutionAgent()
    active_agents = [assignment_solution, r, sokendai, g2, r4, j]

    """ To play in an environment, do something like this... """
    # env = e.Bandit(arms=3)
    # ShellWrapper(env)

    """ To run many agents in an environment, do something like this... """
    def _sne_factory():
        return e.SimpleNavEnvironment(nav_formula)

    # s = competition(_sne_factory, [assignment_solution], steps=100, repeat=5)
    s = competition(_sne_factory, active_agents, steps=100, repeat=1)
    s.description = "negative robustness in " + _sne_factory().codename()
    s.plot(starting_idx=0)
