import numbers


#
# Terms
#
# Terms represent expressions that can be evaluated in a given memory state,
# such as x+1, 2*(x+y), etc.  For falsification, the memory state is given by a
# signal and a time, and the value of the variables is given by their values in
# the signal at that particular time.
# They are given by the gammar:
# t := c | x | t + t | t * t,
# where c is any numeric constant and x any variable name (string).
#


class Term(object):
    def __init__(self):
        self.subterms = []

    def sprint(self):
        return "##undefined term##"

    def __str__(self):
        return self.sprint()

    def __repr__(self):
        return self.sprint()

    def __add__(self, other):
        if not isinstance(other, Term):
            raise TypeError("Cannot add Term and non-Term: {!r}".format(other))
        return Add(self, other)

    def __mul__(self, other):
        if not isinstance(other, Term):
            raise TypeError("Cannot multiply Term and non-Term: {!r}".format(other))
        return Mul(self, other)

    def __eq__(self, other):
        if not isinstance(other, Term):
            return False
        return Equality(self, other)

    def __le__(self, other):
        if not isinstance(other, Term):
            raise TypeError("Cannot compare (<=) Term and non-Term: {!r}".format(other))
        return LTInequality(self, other, strict=False)

    def __lt__(self, other):
        if not isinstance(other, Term):
            raise TypeError("Cannot compare (< ) Term and non-Term: {!r}".format(other))
        return LTInequality(self, other, strict=True)

    def __ge__(self, other):
        if not isinstance(other, Term):
            raise TypeError("Cannot compare (>=) Term and non-Term: {!r}".format(other))
        return GTInequality(self, other, strict=False)

    def __gt__(self, other):
        if not isinstance(other, Term):
            raise TypeError("Cannot compare (> ) Term and non-Term: {!r}".format(other))
        return GTInequality(self, other, strict=True)


class Variable(Term):
    def __init__(self, name):
        super().__init__()
        if isinstance(name, str):
            self.subterms.append(name.strip())
        elif isinstance(name, Variable):
            self.subterms.append(name.subterms[0])
        else:
            raise ValueError("Cannot make a Variable from {!r}.".format(name))

    def sprint(self):
        return self.subterms[0]

    def __repr__(self):
        return self.sprint()

    def __eq__(self, other):
        if isinstance(other, Variable):
            return self.subterms[0] == other.subterms[0]
        else:
            return super().__eq__(other)

    def __hash__(self):
        return hash(str(self))

    def __repr__(self):
        return "Variable('{}')".format(self.subterms[0])


class Constant(Term):
    def __init__(self, val):
        super().__init__()
        if isinstance(val, numbers.Real):
            self.subterms.append(val)

    def sprint(self):
        return str(self.subterms[0])


class Add(Term):
    def __init__(self, *ts):
        super().__init__()
        for t in ts:
            if isinstance(t, Add):
                self.subterms += t.subterms
            elif isinstance(t, Term):
                self.subterms.append(t)

    def sprint(self):
        s = " + "
        if len(self.subterms) > 0:
            return s.join(map(lambda t: "(" + t.sprint() + ")", self.subterms))
        return "0"


class Mul(Term):
    def __init__(self, *ts):
        super().__init__()
        for t in ts:
            if isinstance(t, Mul):
                self.subterms += t.subterms
            elif isinstance(t, Term):
                self.subterms.append(t)

    def sprint(self):
        s = " * "
        if len(self.subterms) > 0:
            return s.join(map(lambda t: "(" + t.sprint() + ")", self.subterms))
        return "1"

#
# Formulas
#
# Formulas are given by the grammar:
# φ := top | t ~ t | ¬ φ | φ ∧ φ | φ U_I φ | φ U φ,
# where t is a term, ~ is =, >, or ≥, and I is an interval [a,b].
# Other connectors, such as t < t, are recovered using syntactic sugar.
#


class Formula(object):
    def __init__(self):
        self.subformulae = []

    def sprint(self):
        return "##undefined formula##"

    def __str__(self):
        return self.sprint()

    def __repr__(self):
        return self.sprint()

    def __and__(self, other):
        if not isinstance(other, Formula):
            raise TypeError("Cannot conjoin Formula and non-Formula {!s}.".format(other))
        return Conjunction(self, other)

    def __invert__(self):
        return Negate(self)


# The following operators are primitive.


class Equality(Formula):
    def __init__(self, l, r):
        super().__init__()
        if isinstance(l, Term) and isinstance(r, Term):
            self.subformulae.append(l)
            self.subformulae.append(r)

    def sprint(self):
        return self.subformulae[0].sprint() + " = " + self.subformulae[1].sprint()


class GTInequality(Formula):
    """ GTInequality(l, r) <==> l > r <==> r is greater than l
    """

    def __init__(self, l, r, strict=True):
        super().__init__()
        if isinstance(l, Term) and isinstance(r, Term):
            self.subformulae.append(l)
            self.subformulae.append(r)
        self.strict = strict

    def sprint(self):
        eq = " " if self.strict else "= "
        return self.subformulae[0].sprint() + " >" + eq + self.subformulae[1].sprint()


class Negate(Formula):
    def __init__(self, other):
        super().__init__()
        if isinstance(other, Formula):
            self.subformulae.append(other)

    def sprint(self):
        return "~ ( " + self.subformulae[0].sprint() + " )"


class Conjunction(Formula):
    def __init__(self, *list_conjuncts):
        super().__init__()
        for conjunct in list_conjuncts:
            if isinstance(conjunct, Conjunction):
                self.subformulae += conjunct.subformulae
            elif isinstance(conjunct, Formula):
                self.subformulae.append(conjunct)

    def sprint(self):
        s = " /\\ "
        if len(self.subformulae) > 0:
            return s.join(map(lambda t: "(" + t.sprint() + ")", self.subformulae))
        return "tt"


class BoundedUntil(Formula):
    def __init__(self, precondition, postcondition, lower, upper):
        super().__init__()
        if isinstance(precondition, Formula) and isinstance(postcondition, Formula):
            self.subformulae.append(precondition)
            self.subformulae.append(postcondition)
        else:
            raise TypeError("Pre- and post-condition of Until must be Formula(s).")
        if lower <= upper:
            self.lower = lower
            self.upper = upper
        else:
            raise ValueError("In range construction, we failed to satisfy {0!s} <= {1!s}.".format(lower, upper))

    def sprint(self):
        return "(" + self.subformulae[0].sprint() + ") U[" + str(self.lower) + ", " + str(self.upper) + "] (" + self.subformulae[1].sprint() + ")"


class Until(Formula):
    def __init__(self, precondition, postcondition):
        super().__init__()
        if isinstance(precondition, Formula) and isinstance(postcondition, Formula):
            self.subformulae.append(precondition)
            self.subformulae.append(postcondition)

    def sprint(self):
        return "(" + self.subformulae[0].sprint() + ") U (" + self.subformulae[1].sprint() + ")"


# The following operators are derived using syntactic sugar.


def LTInequality(l, r, strict=True):
    return GTInequality(r, l, strict=strict)


def Truth():
    return Conjunction([])


def Falsity():
    return Negate(Truth())


def Disjunction(*disjuncts):
    return Negate(Conjunction(*[Negate(_) for _ in disjuncts]))


def Implication(source, target):
    return Disjunction(Negate(source), target)


def BoundedEventually(postcondition, lower, upper):
    return BoundedUntil(Truth(), postcondition, lower, upper)


def BoundedAlways(postcondition, lower, upper):
    return Negate(BoundedEventually(Negate(postcondition), lower, upper))


def Eventually(postcondition):
    return Until(Truth(), postcondition)


def Always(postcondition):
    return Negate(Eventually(Negate(postcondition), lower, upper))


def Interval(name, lower, upper, ls=False, us=False):
    return Conjunction(GTInequality(Variable(name), Constant(lower), strict=ls),
                       GTInequality(Constant(upper), Variable(name), strict=us))


# The following are shorter names for existing operators.


def Var(x):
    return Variable(x)


def Const(c):
    return Constant(c)


def Eq(l, r):
    return Equality(l, r)


def Lt(l, r, strict=True):
    return LTInequality(l, r, strict=strict)


def Not(phi):
    return Negate(phi)


def And(*phis):
    return Conjunction(*phis)


def Or(*phis):
    return Disjunction(*phis)


def Imply(source, target):
    return Implication(source, target)


def E(postcondition):
    return Eventually(postcondition)


def BE(postcondition, lower, upper):
    return BoundedEventually(postcondition, lower, upper)


def BA(postcondition, lower, upper):
    return BoundedAlways(postcondition, lower, upper)


def In(name, lower, upper, ls=False, us=False):
    return Interval(name, lower, upper, ls=ls, us=us)


if __name__ == '__main__':
    x, y, u = Variable('x'), Variable('y'), Variable('u')
    add_test = Variable('x') + Constant(3)
    mul_test = Constant(2.5) * Variable('u')
    comp_test = x < Constant(3)
    compe_test = Constant(3.1415) <= y
    conj_test = comp_test & compe_test

    b, x, y, vx, vy = Var('b'), Var('x'), Var('y'), Var('vx'), Var('vy')
    ineq = (x * x + y * y + vx * vx + vy * vy < Const(0.09)) & (b == Const(3))

    print(str(add_test))
    print(str(mul_test))
    print(str(comp_test))
    print(str(compe_test))
    print(str(conj_test))
    print(str(ineq))
