from operator import mul
from functools import reduce

from stl.formula import *
from stl.timed_signal import Signal


#
# Semantics
#
# Defines the boolean and robustness semantics of formulas.
#


# [term_semantics]: defines the semantics of a term, i.e., its value for a given
# signal and time.
def term_semantics(signal, time, term):
    if not isinstance(term, Term):
        raise TypeError("Term semantics must evaluate a term.")

    # Case analysis on term type.
    if isinstance(term, Variable):
        return signal(time, [term])
    elif isinstance(term, Constant):
        return term.subterms[0]
    elif isinstance(term, Add):
        return sum(term_semantics(signal, time, _) for _ in term.subterms)
    elif isinstance(term, Mul):
        return reduce(mul, [term_semantics(signal, time, _) for _ in term.subterms], 1)

    raise NotImplementedError("term_semantics received a term outside the basic types.")


# [boolean_semantics]: defines the boolean semantics of a formula.
def boolean_semantics(signal, start_time, formula):
    if not isinstance(formula, Formula):
        raise TypeError("Must take semantics of a formula.")
    if not isinstance(signal, Signal):
        raise TypeError("Must take semantics of a signal.")

    # We perform a case analysis on the type of the formula. Most cases are standard...
    if isinstance(formula, Equality):
        term_sems = [term_semantics(signal, start_time, _) for _ in formula.subformulae]
        return term_sems[0] == term_sems[1]
    if isinstance(formula, GTInequality):
        term_sems = [term_semantics(signal, start_time, _) for _ in formula.subformulae]
        if formula.strict:
            return term_sems[0] > term_sems[1]
        else:
            return term_sems[0] >= term_sems[1]
    if isinstance(formula, Negate):
        return not boolean_semantics(signal, start_time, formula.subformulae[0])
    elif isinstance(formula, Conjunction):
        return all([boolean_semantics(signal, start_time, _) for _ in formula.subformulae])

    # ... but the quantifiers in until formulas are discretized
    elif isinstance(formula, BoundedUntil):
        for time in signal.timeRange(start_time, start_time + formula.upper):
            if not boolean_semantics(signal, time, formula.subformulae[0]):
                return False
            if (time >= start_time + formula.lower and
                boolean_semantics(signal, time, formula.subformulae[1])):
                return True
        return False
    elif isinstance(formula, Until):
        for time in signal.timeRange(start_time, signal.upper):
            if not boolean_semantics(signal, time, formula.subformulae[0]):
                return False
            if boolean_semantics(signal, time, formula.subformulae[1]):
                return True
        return False

    raise NotImplementedError("boolean_semantics received a formula outside the basic types.")


# [robustness_semantics]: defines the robustness semantics of a formula.
def robustness_semantics(signal, start_time, formula):
    if not isinstance(formula, Formula):
        raise TypeError("Must take robustness semantics of a formula.")
    if not isinstance(signal, Signal):
        raise TypeError("Must take robustness semantics of a signal.")

    # Assignment, part I: please implement this function to compute robustness semantics of formulae
    # You may wish to reference the boolean_semantics function and use the term_semantics function above.
    # The relevant Formula and Signal objects are defined in formula.py (line 8 - 226) and timed_signal.py (l 8 - 246).
    # You don't need to understand Signal in depth, but its timeRange method and .upper and .lower attributes may be
    # useful. You can also use the test suite below to check your work, but be warned we are intentionally leaving out
    # some tests.
    return 1
    # delete the line above, only here to make the algorithm work.

    raise NotImplementedError("robustness_semantics received a formula outside the basic types.")


def test_boolean_semantics(signal, time, formula, expectation, verbose=False):
    b = boolean_semantics(signal, time, formula)
    if b != expectation:
        print("unexpected result: {!s} @ {!s} |= {!s} ?: {!s}".format(signal, time, formula, b))
    elif verbose:
        print("{!s} @ {!s} |= {!s} ?: {!s}".format(signal, time, formula, b))


def test_robustness_semantics(signal, time, formula, expectation, verbose=False):
    r = robustness_semantics(signal, time, formula)
    if r != expectation:
        print("unexpected result: {!s} @ {!s} |=_r {!s} ?: {!s}".format(signal, time, formula, r))
    elif verbose:
        print("{!s} @ {!s} |=_r {!s} ?: {!s}".format(signal, time, formula, r))


def boolean_semantics_test_suite():
    idSignal = Signal(lambda t: t, upper=5, description='u(t) = t')
    squareSignal = Signal(lambda t: t ** 2, upper=5, description='u(t) = t^2')
    slantSignal = Signal(lambda t: 2 * t - 1, upper=5, description='u(t) = 2t - 1')

    u = Variable('u')
    v = True  # verbosity setting

    for i in range(10):
        assert term_semantics(idSignal, 0.1 * i, Const(2) * u) == 0.2 * i

    tt = Truth()
    test_boolean_semantics(idSignal, 0, tt, True, verbose=v)

    ineq = Const(-2) * u >= Const(0)
    test_boolean_semantics(idSignal, 0, ineq, True, verbose=v)
    test_boolean_semantics(idSignal, 1, ineq, False, verbose=v)

    ineq2 = u > Const(0)
    test_boolean_semantics(idSignal, 0, ineq2, False, verbose=v)
    test_boolean_semantics(idSignal, 1, ineq2, True, verbose=v)

    conj = tt & ineq
    test_boolean_semantics(idSignal, 0, conj, True, verbose=v)
    test_boolean_semantics(idSignal, 1, conj, False, verbose=v)

    conj2 = conj & ineq2
    test_boolean_semantics(idSignal, 0, conj2, False, verbose=v)
    test_boolean_semantics(idSignal, 1, conj2, False, verbose=v)

    neg = ~ conj2
    test_boolean_semantics(idSignal, 0, neg, True, verbose=v)
    test_boolean_semantics(idSignal, 1, neg, True, verbose=v)

    ff = Falsity()
    test_boolean_semantics(idSignal, 1, ff, False, verbose=v)

    ltineq = u < Const(0)
    test_boolean_semantics(idSignal, 0, ltineq, False, verbose=v)
    test_boolean_semantics(idSignal, 1, ltineq, False, verbose=v)

    ev_true = BoundedEventually(u >= Const(3), 2, 4)
    test_boolean_semantics(idSignal, 0, ev_true, True, verbose=v)

    ev_false = BoundedEventually(u >= Const(5), 2, 4)
    test_boolean_semantics(idSignal, 0, ev_false, False, verbose=v)


def robustness_semantics_test_suite():
    idSignal = Signal(lambda t: t, upper=5, description='u(t) = t')
    squareSignal = Signal(lambda t: t ** 2, upper=5, description='u(t) = t^2')
    slantSignal = Signal(lambda t: 2 * t - 1, upper=5, description='u(t) = 2t - 1')

    u = Variable('u')
    v = True  # verbosity setting

    for i in range(10):
        assert term_semantics(idSignal, 0.1 * i, Const(2) * u) == 0.2 * i

    tt = Truth()
    test_robustness_semantics(idSignal, 0, tt, float('inf'), verbose=v)

    ineq = Const(-2) * u >= Const(0)
    test_robustness_semantics(idSignal, 0, ineq, 0, verbose=v)
    test_robustness_semantics(idSignal, 1, ineq, -2, verbose=v)

    ineq2 = u > Const(0)
    test_robustness_semantics(idSignal, 0, ineq2, 0, verbose=v)
    test_robustness_semantics(idSignal, 1, ineq2, 1, verbose=v)

    conj = tt & ineq
    test_robustness_semantics(idSignal, 0, conj, 0, verbose=v)
    test_robustness_semantics(idSignal, 1, conj, -2, verbose=v)

    conj2 = conj & ineq2
    test_robustness_semantics(idSignal, 0, conj2, 0, verbose=v)
    test_robustness_semantics(idSignal, 1, conj2, -2, verbose=v)

    neg = ~ conj2
    test_robustness_semantics(idSignal, 0, neg, 0, verbose=v)
    test_robustness_semantics(idSignal, 1, neg, 2, verbose=v)

    ff = Falsity()
    test_robustness_semantics(idSignal, 1, ff, float('-inf'), verbose=v)

    ltineq = u < Const(0)
    test_robustness_semantics(idSignal, 0, ltineq, 0, verbose=v)
    test_robustness_semantics(idSignal, 1, ltineq, -1, verbose=v)

    ev_true = BoundedEventually(u >= Const(3), 2, 4)
    test_robustness_semantics(idSignal, 0, ev_true, 1, verbose=v)

    ev_false = BoundedEventually(u >= Const(5), 2, 4)
    test_robustness_semantics(idSignal, 0, ev_false, -1, verbose=v)


if __name__ == '__main__':
    robustness_semantics_test_suite()
