import math
import numbers
import matplotlib.pyplot as plt
import numpy as np

from stl.formula import Variable


#
# Signal
#
# Implements signals.
#


class Signal:
    """ A Signal represents a continuous time signal over a bounded interval.
    It can be constructed from a function or an iterable and then also used as a function or an iterable."""
    def __init__(self, generator, variables=('u',), lower=0, upper=1, delta=1E-4, description=None, array_list=None):
        self.lower = lower
        self.upper = upper
        self.delta = delta
        self.callable = generator if callable(generator) else None
        self.arrays = array_list
        self.description = description
        self.times = self.timeRange(self.lower, self.upper)

        # First we find the dimension of the signal. This can be determined from the generator, array_dict, or
        # variable name list. We have to be sure these are consistent.
        if self.callable:
            try:
                self.dim = len(self.callable(lower))
            except TypeError:  # triggered when the generator produces a non-iterable
                self.dim = 1
                self.callable = lambda _: (generator(_),)
        else:
            generator = list(generator)
            try:
                self.dim = len(generator[0])
            except:
                self.dim = 1
        if array_list is not None and self.dim != len(array_list):
            raise ValueError("Array dictionary for signal has {!s} values, Signal has dimension {!s}."
                             .format(len(array_list), self.dim))

        # If variables is the default value, we may expand it to fit the dimension. Otherwise we have an error.
        if variables == ('u',) and self.dim > 1:
            variables = ['u' + str(k + 1) for k in range(self.dim)]
        if self.dim != len(variables):
            raise ValueError("Variables for signal has {!s} values, Signal has dimension {!s}."
                             .format(len(variables), self.dim))

        # We can now set the list of Variables for our Signal.
        self.variables = [Variable(var) for var in variables]

        # Self.arrays stores arrays consisting of the values of the variables in the signal at the times in self.times.
        # This is a dictionary of the form {var1: [val1, val2, ..., valn], ..., vark: [...]}. (n = len(self.times))
        # When this dictionary is provided in the constructor (via array_dict), we do some validation to make sure it
        # fits with the rest of the data provided. Otherwise, we have to generate this dictionary from the generator.
        if array_list is not None:
            for var_idx, vals in enumerate(array_list):
                if len(vals) != len(self.times):
                    var = self.variables[var_idx]
                    raise ValueError("Array for variable {!s} has {!s} values, should have {!s}".format(var, len(vals), len(self.times)))
        else:
            self.arrays = [[None] * len(self.times) for _ in range(self.dim)]
            for t_idx, t in enumerate(self.times):
                if self.callable:
                    vals = self.callable(t)
                else:
                    vals = generator[t_idx]
                for var_idx in range(self.dim):
                    self.arrays[var_idx][t_idx] = vals[var_idx]

    def __len__(self):
        return self.upper - self.lower

    def timeRange(self, lower, upper):
        return np.linspace(lower, upper, num=round((upper - lower) / self.delta + 1))

    def __iter__(self):
        if self.dim == 1:
            return self.arrays[0]
        else:
            return list(zip(*self.arrays))

    def __getitem__(self, item):
        if self.dim == 1:
            return self.arrays[0][item]
        else:
            return tuple(self.arrays[d][item] for d in range(self.dim))

    def __str__(self):
        return self.description or "Undescribed Signal"

    def __call__(self, time, variables=None, unpack_singleton=True):
        if time < self.lower or time > self.upper:
            raise ValueError("Signal called at {!s}, which is outside domain of definition [{!s}, {!s}]."
                             .format(time, self.lower, self.upper))

        if self.callable:
            value = self.callable(time)
        else:
            idx = math.floor(time / self.delta)
            while idx > 0 and self.times[idx] > time:
                idx -= 1
            while idx + 1 < len(self.times) and self.times[idx] + self.delta <= time:
                idx += 1
            value = self[idx]

        return self._project(value, variables, unpack_singleton)

    def _project_idxs(self, tup, idxs=None, unpack_singleton=True):
        if not idxs:
            ret = tup
        else:
            if isinstance(tup, numbers.Real):
                tup = (tup,)
            if max(idxs) >= len(tup):
                raise ValueError("Signal projection: max index {!s} greater than tuple size {!s}.".format(max(idxs), len(tup)))
            ret = tuple([tup[i] for i in idxs])

        if hasattr(ret, '__len__') and len(ret) == 1 and unpack_singleton:
            ret = ret[0]
        return ret

    def _idxs_from_variables(self, variables=None):
        if not variables:
            return None

        var_list = [Variable(_) for _ in variables]
        idx_list = [i for i in range(len(self.variables)) if self.variables[i] in var_list]
        if not idx_list:
            raise ValueError("Empty Signal projection.")
        return idx_list

    def _project(self, tup, variables=None, unpack_singleton=True):
        idxs = self._idxs_from_variables(variables)
        return self._project_idxs(tup, idxs=idxs, unpack_singleton=unpack_singleton)

    def project(self, variables=None, unpack_singleton=True):
        if not variables:
            return self

        description = " of " + str(self)
        if len(variables) == 1:
            description = str(variables[0]) + description
        else:
            description = 'projection' + description

        new_array_list = self._project(self.arrays, variables, False)

        if self.callable:
            new_call = lambda t: self(t, variables=variables, unpack_singleton=unpack_singleton)
            return Signal(new_call, variables, self.lower, self.upper, self.delta, description=description, array_list=new_array_list)
        else:
            good_tuple = self._project(self[0], variables=variables, unpack_singleton=False)
            return Signal(good_tuple, variables, self.lower, self.upper, self.delta, description=description, array_list=new_array_list)

    def pair(self, other):
        if not isinstance(other, Signal):
            raise ValueError("Cannot pair Signal {!r} and non-Signal {!r}.".format(self, other))
        if self.lower != other.lower or self.upper != other.upper or self.delta != other.delta:
            raise ValueError("Cannot pair Signals defined on different intervals.")

        good_tuple = [np.zeros(self.dim + other.dim)]  # just for passing dimension check...
        new_array = np.concatenate((np.array(self.arrays), np.array(other.arrays)), axis=0)
        new_vars = self.variables + other.variables

        return Signal(good_tuple, new_vars, self.lower, self.upper, self.delta, array_list=new_array)

    def __add__(self, other):
        if not isinstance(other, Signal):
            raise ValueError("Cannot add Signal {!r} and non-Signal {!r}.".format(self, other))
        if self.dim != other.dim or self.variables != other.variables or self.delta != other.delta:
            raise ValueError("Cannot add Signals of unlike type.")
        if self.lower != other.lower or self.upper != other.upper:
            raise ValueError("Cannot add Signals defined on different intervals.")

        # s = lambda t: tuple([self(t, unpack_singleton=False)[i] + other(t, unpack_singleton=False)[i] for i in range(self.dim)])
        good_tuple = [np.zeros(self.dim)]  # just for passing dimension check...
        new_array = np.array(self.arrays) + np.array(other.arrays)

        return Signal(good_tuple, self.variables, self.lower, self.upper, self.delta, array_list=new_array)

    def __mul__(self, other):
        if not isinstance(other, Signal):
            raise ValueError("Cannot multiply Signal {!r} and non-Signal {!r}.".format(self, other))
        if self.dim != other.dim or self.variables != other.variables or self.delta != other.delta:
            raise ValueError("Cannot multiply Signals of unlike type.")
        if self.lower != other.lower or self.upper != other.upper:
            raise ValueError("Cannot multiply Signals defined on different intervals.")

        # p = lambda t: tuple([self(t, unpack_singleton=False)[i] * other(t, unpack_singleton=False)[i] for i in range(self.dim)])
        good_tuple = [np.zeros(self.dim)]  # just for passing dimension check...
        new_array = np.array(self.arrays) * np.array(other.arrays)
        return Signal(good_tuple, self.variables, self.lower, self.upper, self.delta, array_list=new_array)

    def __rshift__(self, other):
        """ Concatenation of signals. """
        if isinstance(other, Signal):
            if self.dim != other.dim or self.variables != other.variables or self.delta != other.delta:
                raise ValueError("Cannot concatenate Signals of unlike type.")

            def u(t):
                if t <= self.upper:
                    return self(t)
                else:
                    shifted_t = t - self.upper + other.lower
                    return other(shifted_t)

            return Signal(u, self.variables, self.lower, self.upper - other.lower + other.upper,
                          description="concatenation of {!s} and {!s}".format(self, other))

        if isinstance(other, tuple):
            if self.dim  == len(other):
                self.upper += self.delta
                self.times = np.append(self.times, [self.upper])
                for idx in range(self.dim):
                    self.arrays[idx] = list(self.arrays[idx])
                    self.arrays[idx].append(other[idx])
                self.callable = None
                return self

        if self.dim == 1:
            self.upper += self.delta
            self.times = np.append(self.times, [self.upper])
            self.arrays[0] = list(self.arrays[0])
            self.arrays[0].append(other[0])
            self.callable = None
            return self

        raise TypeError("Signal concatenation failed.")

    def plot(self, variables=None, starting_idx=0, colormap=None, show=True, save_to=None):
        COLORS = ['b', 'g', 'r', 'c', 'm', 'y', 'k']
        if not variables:
            variables = self.variables
        if not colormap:
            colormap = {}

        fig, ax = plt.subplots()
        for idx, variable in enumerate(variables):
            data = self.project(variables=(Variable(variable),)).arrays[0][starting_idx:]
            if variable in colormap:
                ax.plot(self.times[starting_idx:], data, colormap[variable], label=variable)
            else:
                ax.plot(self.times[starting_idx:], data, COLORS[idx % len(COLORS)], label=variable)
        ax.set(xlabel='time step', title=str(self))
        ax.grid()
        if len(variables) > 1:
            plt.legend()

        if save_to:
            fig.savefig(save_to)
        if show:
            plt.show()


def average_signals(*sgs, description=None):
    if not sgs:
        raise ValueError("Cannot take average of empty list of Signals.")

    first = sgs[0]
    factor_sig = Signal(lambda t: (1.0/len(sgs),) * first.dim, first.variables, first.lower, first.upper, first.delta)
    avg = first * factor_sig
    for sig in sgs[1:]:
        avg = avg + sig * factor_sig
    if description:
        avg.description = description
    return avg


def pair_signals(*sgs):
    if not sgs:
        raise ValueError("Cannot pair up empty list of Signals.")

    ret = sgs[0]
    for sig in sgs[1:]:
        ret = ret.pair(sig)
    return ret


if __name__ == "__main__":
    s2 = Signal(lambda t: (t, t + 2), variables=('x', 'y'), description='skis')
    x, y = Variable('x'), Variable('y')
    print(s2(0.50, variables=[x]))
    print(s2(0.50, variables=[y]))
    s2.plot(['x', 'y'])

    projection = s2.project([x])
    projection.plot()

    from math import cos

    c = Signal(cos, description='cosine')
    c.plot()

