from math import sqrt, log
from random import randrange, uniform, normalvariate, expovariate


from utils import insert_sorted_list
from falsification.point import Point


#
# Optimisation
#
# A class implementing optimisation methods.
# An object of this class has:
# - an [initialise] method that is called at the beginning of the falsification
#   process to initialise the global structure used by the algorithm,
# - a [generate] function that is called multiple times throughout the
#   falsification process to generate new inputs (positions or signals) to try,
# - an [update] function that is called after the robustness of the output is
#   known, to update the structure.
#


class Optimisation(object):

    def __init__(self, initialise, generate, update):
        """Creates an optimisation method:
        - initialise: how to initialise the global falsification structre
        - generate: how to generate new positions/signals,
        - update: how to update the global falsification structure.
        """
        self.initialise = initialise
        self.generate   = generate
        self.update     = update


#
# Random generation
#
# This "optimisation" method chooses a new point at random each time it is
# called.
#


def random_initialise(struct):
    pass


def random_generate(struct):
    q = struct["mode"]
    n = struct["number-modes"]
    variables = struct["variables"]
    invariant = struct["invariant"]

    new = {}
    if q != None:
        new[q] = randrange(n)
    for v in variables:
        (a, b) = invariant(q, v)
        if a == float("-inf"):
            if b == float("inf"):
                new[v] = normalvariate(0, 1)
            else:
                new[v] = b - expovariate(1)
        elif b == float("inf"):
            new[v] = a + expovariate(1)
        else:
            new[v] = uniform(a, b)
    return new


def random_update(position, robustness, struct):
    pass


random = Optimisation(random_initialise,
                      random_generate,
                      random_update)


#
# Biased random generation
#
# This optimisation method chooses a new point at random each time it is
# called, but chooses the mode using a balance of exploration and exploitation.
# This balance is achieved using a factor [lambda] (a bigger lambda gives more
# weight to exploitation).
#


def random_biased_initialise(lambda_, struct):
    if struct["mode"] != None:
        struct["for-optimisation"]["lambda"] = lambda_
        # [biases]: stores for each mode:
        # - the minimum robustness of outputs from inputs in this mode (used to
        #   compute an exploitation score),
        # - the number of inputs in this tried (used to compute an exploitation
        #   score).
        struct["for-optimisation"]["biases"] = ([[]] * struct["number-modes"])
        for i in range(struct["number-modes"]):
            struct["for-optimisation"]["biases"][i] = [float("inf"),0]


def random_biased_generate_mode(struct):
    attractiveness = [0] * struct["number-modes"]
    for i in range(struct["number-modes"]):
        if struct["for-optimisation"]["biases"][i][1] == 0:
            return i
        # exploitation
        attractiveness[i] += 1 / ((struct["for-optimisation"]["biases"][i][0])
                                  ** struct["for-optimisation"]["lambda"])
        # exploration
        attractiveness[i] += sqrt(2*log(len(struct["tried"])) /
                                  struct["for-optimisation"]["biases"][i][1])
    res = 0
    for i in range(1,struct["number-modes"]):
        if attractiveness[i] > attractiveness[res]:
            res = i
    return res


def random_biased_generate_in_mode(n, struct):
    q = struct["mode"]
    variables = struct["variables"]
    invariant = struct["invariant"]

    new = {q: n}
    for v in variables:
        (a, b) = invariant(n, v)
        if a == float("-inf"):
            if b == float("inf"):
                new[v] = normalvariate(0, 1)
            else:
                new[v] = b - expovariate(1)
        elif b == float("inf"):
            new[v] = a + expovariate(1)
        else:
            new[v] = uniform(a, b)
    return new


def random_biased_generate(struct):
    if struct["mode"] == None:
        return random_generate(struct)
    q = random_biased_generate_mode(struct)
    p = random_biased_generate_in_mode(q, struct)
    return p


def random_biased_update(p, r, struct):
    q = struct["mode"]
    if q != None:
        # we have visited p's mode once more
        struct["for-optimisation"]["biases"][p[q]][1] += 1
        # if we get a better robustness, we update it
        if r < struct["for-optimisation"]["biases"][p[q]][0]:
            struct["for-optimisation"]["biases"][p[q]][0] = r


def random_biased(lambda_):
    def random_biased_initisalise_complete(struct):
        random_biased_initialise(lambda_, struct)

    return Optimisation(random_biased_initisalise_complete,
                        random_biased_generate,
                        random_biased_update)


#
# Nelder-Mead
#
# This optimisation method keeps a simplex structure as it tries to approach a
# minimum.  This simplex is modified by simple geometric operations that drive
# it towards a minimum.  Sometimes, a new simplex is generated from scratch to
# avoid falling into local minima.  The method is based on four parameters:
# - [alpha]: the reflexion parameter, a bigger alpha means vertices are
#   reflected farther, should be > 0,
# - [beta]: the contraction parameter, a bigger beta means vertices are
#   contracted closer to their original position, should be > 0 and < 1,
# - [gamma]: the expansion parameter, a bigger gamma means vertices are expanded
#   farther, should be > 0,
# - [epsilon]: the threshold to know when to generate a new simplex (a bigger
#   epsilon means more time is spent on each simplex), should be > 0.
# It also needs a function [generate_simplex] to generate vertices for new
# simplices.
#


def nelder_mead_initialise(alpha, beta, gamma, epsilon, generate_simplex,
                           struct):
    if alpha <= 0:
        raise ValueError("In Nelder-Mead, alpha must be > 0.")
    if beta <= 0 or beta >= 1:
        raise ValueError("In Nelder-Mead, beta must be > 0 and < 1.")
    if gamma <= 0:
        raise ValueError("In Nelder-Mead, gamma must be > 0.")
    if epsilon <= 0:
        raise ValueError("In Nelder-Mead, epsilon must be > 0.")
    struct["for-optimisation"]["alpha"]   = alpha
    struct["for-optimisation"]["beta"]    = beta
    struct["for-optimisation"]["gamma"]   = gamma
    struct["for-optimisation"]["epsilon"] = epsilon
    # [step]:
    # - 0: vertices for a new simplex (also start of the algorithm)
    # - 1: reflection
    # - 2: expansion
    # - 3: contraction
    # - 4: contracting simplex
    # attention:
    # - in generate: next step to do (e.g., at step 1, we need to compute
    #   reflection)
    # - in update: step just done (e.g., at step 1, we have just computed
    #   reflection)
    struct["for-optimisation"]["step"] = 0
    struct["for-optimisation"]["simplex"] = []
    # [memory]: maps variable names used in the algorithm to their values
    struct["for-optimisation"]["memory"] = {}
    # [simplex]: stores the simplex once it is compeletely computed
    struct["struct"] = {"mode"            : struct["mode"],
                        "number-modes"    : struct["number-modes"],
                        "variables"       : struct["variables"],
                        "invariant"       : struct["invariant"],
                        "optimisation"    : struct["optimisation"],
                        "time_sim"        : struct["time_sim"],
                        "best"            : (None, float("inf")),
                        "tried"           : struct["tried"],
                        "for-optimisation": {},
                        "time"            : 0,
                        "simplex"         : []}
    generate_simplex.initialise(struct["struct"])


def nelder_mead_generate(generate_simplex, struct):
    n = len(struct["variables"])
    memory = struct["for-optimisation"]["memory"]
    # vertices for a new simplex (also start of the algorithm)
    if struct["for-optimisation"]["step"] == 0:
        return generate_simplex.generate(struct["struct"])
    # reflection
    if struct["for-optimisation"]["step"] == 1:
        memory["p"] = struct["for-optimisation"]["simplex"][:-1]
        memory["h"] = struct["for-optimisation"]["simplex"][-1]
        zero = {}
        for i in memory["p"][0][0].get_coordinates():
            zero[i] = 0
        memory["centroid"] = Point(zero, lambda x: x != struct["mode"])
        memory["centroid"][struct["mode"]] = memory["p"][0][0][struct["mode"]]

        # Assignment, part II: please implement the rest of this function.
        # Everything you need to implement can is akin to the nelder_mead_loop
        # function in nelder_mead.py.  The part of the function that differs
        # from it (because it does picks a new simplex when some quantity falls
        # below a certain threshold) is already written (steps 0 and 4), as well
        # as the begining of step 1, to give you a feel of how to write this
        # function.  It may be valuable to spend some time to get some
        # superficial understanding of what is written in the rest of the
        # function, as well as in nelder_mead_update.
        return random.generate()
        # delete the line above, only here to make the algorithm work.

    # expansion
    elif struct["for-optimisation"]["step"] == 2:

        # Complete here too.
        return random.generate()
        # delete the line above, only here to make the algorithm work.

    # contraction
    elif struct["for-optimisation"]["step"] == 3:

        # Complete here too.
        return random.generate()
        # delete the line above, only here to make the algorithm work.

    # contracting simplex
    elif struct["for-optimisation"]["step"] == 4:
        return (((memory["p"][memory["i"]][0] + memory["p"][0][0]) / 2).
                get_coordinates())

    raise ValueError("Unknown step in Nelder-Mead algorithm: {}.".
                     format(step))


def nelder_mead_stop(struct):
    n = len(struct["variables"])
    simplex = struct["for-optimisation"]["simplex"]
    if len(simplex) <= n:
        return False
    average = (sum(list(map(lambda x: x[1], simplex))) / (n+1))
    x = sqrt(sum(list(map(lambda x: (x[1]-average) ** 2, simplex))))
    return x < struct["for-optimisation"]["epsilon"]


def nelder_mead_update(p, r, generate_simplex, struct):
    n = len(struct["variables"])
    memory = struct["for-optimisation"]["memory"]
    # we generated a new vertex of the initial simplex
    if struct["for-optimisation"]["step"] == 0:
        generate_simplex.update(p, r, struct["struct"])
        # if the simplex is complete
        if len(struct["struct"]["simplex"]) == n+1:
            struct["for-optimisation"]["simplex"] = struct["struct"]["simplex"]
            struct["struct"]["simplex"] = []
            struct["for-optimisation"]["step"] = 1
            memory = {}
            # if the simplex is too "flat" (all values are close to each other),
            # create a new simplex
            if nelder_mead_stop(struct):
                struct["for-optimisation"]["step"] = 0
    # we computed the reflection
    elif struct["for-optimisation"]["step"] == 1:
        memory["yr"] = r
        # if it's better than all vertices, try expanding
        if r < memory["p"][0][1]:
            struct["for-optimisation"]["step"] = 2
            return
        # if it's not the worst, use it as a new vertex
        elif r <= memory["p"][-1][1]:
            p = Point(p, lambda x: x != struct["mode"])
            insert_sorted_list((p, r), memory["p"], lambda x: x[1])
            struct["for-optimisation"]["simplex"] = memory["p"]
            # struct["for-optimisation"]["step"] = 1 # not needed because it's already true
            if nelder_mead_stop(struct):
                struct["for-optimisation"]["step"] = 0
            return
        # otherwise, try contracting
        memory["ymin"] = memory["yr"]
        if memory["yr"] > memory["h"][1]:
            memory["ymin"] = memory["h"][1]
            memory["h"] = (memory["refl"], memory["yr"])
        struct["for-optimisation"]["step"] = 3
    # we computed the expansion
    elif struct["for-optimisation"]["step"] == 2:
        # if it's better than all vertices, use it as a new vertex
        if r < memory["p"][0][1]:
            p = Point(p, lambda x: x != struct["mode"])
            insert_sorted_list((p, r), memory["p"], lambda x: x[1])
            struct["for-optimisation"]["simplex"] = memory["p"]
            struct["for-optimisation"]["step"] = 1 # and restart from computing the reflection
            if nelder_mead_stop(struct): # except if the simplex is too flat
                struct["for-optimisation"]["step"] = 0
            return
        # otherwise, forget about the expansion and do as in step 1
        if memory["yr"] <= memory["p"][-1][1]:
            p = Point(p, lambda x: x != struct["mode"])
            insert_sorted_list((p, memory["yr"]), memory["p"],
                                lambda x: x[1])
            struct["for-optimisation"]["simplex"] = memory["p"]
            struct["for-optimisation"]["step"] = 1
            if nelder_mead_stop(struct):
                struct["for-optimisation"]["step"] = 0
            return
        memory["ymin"] = memory["yr"]
        if memory["yr"] > memory["h"][1]:
            memory["ymin"] = memory["h"][1]
            memory["h"] = (memory["refl"], memory["yr"])
        struct["for-optimisation"]["step"] = 3
    # we computed the contraction
    elif struct["for-optimisation"]["step"] == 3:
        # if the contraction is worse than the previous vertices, contract the
        # whole simplex
        if r > memory["ymin"]:
            memory["p"].append(memory["h"])
            memory["i"] = 1
            memory["simplex'"] = [struct["for-optimisation"]["simplex"][0]]
            struct["for-optimisation"]["step"] = 4
            return
        p = Point(p, lambda x: x != struct["mode"])
        insert_sorted_list((p, r), memory["p"], lambda x: x[1])
        struct["for-optimisation"]["simplex"] = memory["p"]
        struct["for-optimisation"]["step"] = 1
    # we computed one vertex for the simplex
    elif struct["for-optimisation"]["step"] == 4:
        # if the vertex is valid, add it to the contracted simplex
        if r < float("inf"):
            memory["i"] += 1
            p = Point(p, lambda x: x != struct["mode"])
            insert_sorted_list((p, r), memory["simplex'"], lambda x: x[1])
            if memory["i"] == len(struct["for-optimisation"]["simplex"]):
                struct["for-optimisation"]["simplex"] = memory["simplex'"]
                struct["for-optimisation"]["step"] = 1
                if nelder_mead_stop(struct):
                    struct["for-optimisation"]["step"] = 0
        else:
            struct["for-optimisation"]["simplex"] = []
            struct["for-optimisation"]["step"] = 0

    if not struct["for-optimisation"]["step"] in [0, 1, 2, 3, 4]:
        raise ValueError("Unknown step in Nelder-Mead algorithm: {}.".
                         format(struct["for-optimisation"]["step"]))


def nelder_mead(alpha, beta, gamma, epsilon, generate_simplex):
    def nelder_mead_initialise_abce(struct):
        nelder_mead_initialise(alpha, beta, gamma, epsilon, generate_simplex,
                               struct)

    def nelder_mead_generate_complete(struct):
        return nelder_mead_generate(generate_simplex, struct)

    def nelder_mead_update_complete(p, r, struct):
        nelder_mead_update(p, r, generate_simplex, struct)

    return Optimisation(nelder_mead_initialise_abce,
                        nelder_mead_generate_complete,
                        nelder_mead_update_complete)


#
# Nelder-Mead with biased random sampling for simplex vertices
#

def nelder_mead_random_biased_simplex_initialise(lambda_, struct):
    if struct["mode"] != None:
        struct["for-optimisation"]["lambda"] = lambda_
        # [biases]: stores for each mode:
        # - the minimum robustness of outputs from inputs in this mode (used to
        #   compute an exploitation score),
        # - the number of inputs in this tried (used to compute an exploitation
        #   score).
        struct["for-optimisation"]["biases"] = ([[]] * struct["number-modes"])
        for i in range(struct["number-modes"]):
            struct["for-optimisation"]["biases"][i] = [float("inf"),0]
        # new for Nelder-Mead
        struct["for-optimisation"]["simplices"] = []
        for i in range(struct["number-modes"]):
            struct["for-optimisation"]["simplices"].append([])
    # new for Nelder-Mead
    else:
        struct["for-optimisation"]["simplex"] = []


# nelder_mead_random_biased_simplex_generate = random_biased_generate


def nelder_mead_random_biased_simplex_update(p, r, struct):
    n = len(struct["variables"])
    q = struct["mode"]
    if q != None:
        # we have visited p's mode once more
        struct["for-optimisation"]["biases"][p[q]][1] += 1
        # if we get a better robustness, we update it
        if r < struct["for-optimisation"]["biases"][p[q]][0]:
            struct["for-optimisation"]["biases"][p[q]][0] = r
        # new for Nelder-Mead
        if r < float("inf"):
            insert_sorted_list((Point(p, lambda x: x != struct["mode"]), r),
                               struct["for-optimisation"]["simplices"][p[q]],
                               lambda x: x[1])
            if len(struct["for-optimisation"]["simplices"][p[q]]) == n+1:
                struct["simplex"] = struct["for-optimisation"]["simplices"][p[q]]
                struct["for-optimisation"]["simplices"][p[q]] = []
    # new for Nelder-Mead
    else:
        if r < float("inf"):
            insert_sorted_list((p,r), struct["for-optimisation"]["simplex"],
                               lambda x: x[1])
            if len(struct["for-optimisation"]["simplex"]) == n+1:
                struct["simplex"] = struct["for-optimisation"]["simplex"]
                struct["for-optimisation"]["simplex"] = []


def nelder_mead_random_biased(alpha, beta, gamma, epsilon, lambda_):
    def simplex_initialise(struct):
        nelder_mead_random_biased_simplex_initialise(lambda_, struct)

    def simplex_generate(struct):
        return nelder_mead_random_biased_simplex_generate(struct)

    generate_simplex = Optimisation(simplex_initialise,
                                    random_biased_generate,
                                    nelder_mead_random_biased_simplex_update)

    return nelder_mead(alpha, beta, gamma, epsilon, generate_simplex)
